import importlib
import os
import pkgutil
import sys
import typing as t
from collections import defaultdict
from typing import OrderedDict

import click
import yaml
from tabulate import tabulate

from clarifai.utils.logging import logger


def from_yaml(filename: str):
  try:
    with open(filename, 'r') as f:
      return yaml.safe_load(f)
  except yaml.YAMLError as e:
    click.echo(f"Error reading YAML file: {e}", err=True)
    return {}


def dump_yaml(data, filename: str):
  try:
    with open(filename, 'w') as f:
      yaml.dump(data, f)
  except Exception as e:
    click.echo(f"Error writing YAML file: {e}", err=True)


# Dynamically find and import all command modules from the cli directory
def load_command_modules():
  package_dir = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'cli')

  for _, module_name, _ in pkgutil.iter_modules([package_dir]):
    if module_name not in ['base', '__main__']:  # Skip the base.py and __main__ file itself
      importlib.import_module(f'clarifai.cli.{module_name}')


def display_co_resources(response,
                         custom_columns={
                             'ID': lambda c: c.id,
                             'USER_ID': lambda c: c.user_id,
                             'DESCRIPTION': lambda c: c.description,
                         }):
  """Display compute orchestration resources listing results using rich."""

  formatter = TableFormatter(custom_columns)
  print(formatter.format(list(response), fmt="plain"))


class TableFormatter:

  def __init__(self, custom_columns: OrderedDict):
    """
        Initializes the TableFormatter with column headers and custom column mappings.

        :param headers: List of column headers for the table.
        """
    self.custom_columns = custom_columns

  def format(self, objects, fmt='plain'):
    """
        Formats a list of objects into a table with custom columns.

        :param objects: List of objects to format into a table.
        :return: A string representing the table.
        """
    # Prepare the rows by applying the custom column functions to each object
    rows = []
    for obj in objects:
      #   row = [self.custom_columns[header](obj) for header in self.headers]
      row = [f(obj) for f in self.custom_columns.values()]
      rows.append(row)

    # Create the table
    table = tabulate(rows, headers=self.custom_columns.keys(), tablefmt=fmt)
    return table


class AliasedGroup(click.Group):

  def __init__(self,
               name: t.Optional[str] = None,
               commands: t.Optional[t.Union[t.MutableMapping[str, click.Command], t.Sequence[
                   click.Command]]] = None,
               **attrs: t.Any) -> None:
    super().__init__(name, commands, **attrs)
    self.alias_map = {}
    self.command_to_aliases = defaultdict(list)

  def add_alias(self, cmd: click.Command, alias: str) -> None:
    self.alias_map[alias] = cmd
    if alias != cmd.name:
      self.command_to_aliases[cmd].append(alias)

  def command(self, aliases=None, *args,
              **kwargs) -> t.Callable[[t.Callable[..., t.Any]], click.Command]:
    cmd_decorator = super().command(*args, **kwargs)
    if aliases is None:
      aliases = []

    def aliased_decorator(f):
      cmd = cmd_decorator(f)
      if cmd.name:
        self.add_alias(cmd, cmd.name)
      for alias in aliases:
        self.add_alias(cmd, alias)
      return cmd

    f = None
    if args and callable(args[0]):
      (f,) = args
    if f is not None:
      return aliased_decorator(f)
    return aliased_decorator

  def group(self, aliases=None, *args,
            **kwargs) -> t.Callable[[t.Callable[..., t.Any]], click.Group]:
    cmd_decorator = super().group(*args, **kwargs)
    if aliases is None:
      aliases = []

    def aliased_decorator(f):
      cmd = cmd_decorator(f)
      if cmd.name:
        self.add_alias(cmd, cmd.name)
      for alias in aliases:
        self.add_alias(cmd, alias)
      return cmd

    f = None
    if args and callable(args[0]):
      (f,) = args
    if f is not None:
      return aliased_decorator(f)
    return aliased_decorator

  def get_command(self, ctx: click.Context, cmd_name: str) -> t.Optional[click.Command]:
    rv = click.Group.get_command(self, ctx, cmd_name)
    if rv is not None:
      return rv
    return self.alias_map.get(cmd_name)

  def format_commands(self, ctx, formatter):
    sub_commands = self.list_commands(ctx)

    rows = []
    for sub_command in sub_commands:
      cmd = self.get_command(ctx, sub_command)
      if cmd is None or getattr(cmd, 'hidden', False):
        continue
      if cmd in self.command_to_aliases:
        aliases = ', '.join(self.command_to_aliases[cmd])
        sub_command = f'{sub_command} ({aliases})'
      cmd_help = cmd.help
      rows.append((sub_command, cmd_help))

    if rows:
      with formatter.section("Commands"):
        formatter.write_dl(rows)


def validate_context(ctx):
  if ctx.obj == {}:
    logger.error("CLI config file missing. Run `clarifai login` to set up the CLI config.")
    sys.exit(1)



from typing import Dict, Generator

from clarifai_grpc.grpc.api import resources_pb2, service_pb2

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.errors import UserError
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import logger


class Module(Lister, BaseClient):
  """Module is a class that provides access to Clarifai API endpoints related to Module information."""

  def __init__(self,
               url: str = None,
               module_id: str = None,
               module_version: Dict = {'id': ""},
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes a Module object.

        Args:
            url (str): The URL to initialize the module object.
            module_id (str): The Module ID to interact with.
            module_version (dict): The Module Version to interact with.
            base_url (str): Base API url. Default "https://api.clarifai.com"
            pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT.
            token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN.
            root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
            **kwargs: Additional keyword arguments to be passed to the Module.
        """
    if url and module_id:
      raise UserError("You can only specify one of url or module_id.")
    if not url and not module_id:
      raise UserError("You must specify one of url or module_id.")
    if url:
      user_id, app_id, module_id, module_version_id = ClarifaiUrlHelper.split_module_ui_url(url)
      module_version = {'id': module_version_id}
      kwargs = {'user_id': user_id, 'app_id': app_id}

    self.kwargs = {**kwargs, 'id': module_id, 'module_version': module_version}
    self.module_info = resources_pb2.Module(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=self.user_id,
        app_id=self.app_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  def list_versions(self, page_no: int = None,
                    per_page: int = None) -> Generator['Module', None, None]:
    """Lists all the module versions for the module.

    Args:
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Moudle: Module objects for versions of the module.

    Example:
        >>> from clarifai.client.module import Module
        >>> module = Module(module_id='module_id', user_id='user_id', app_id='app_id')
        >>> all_Module_versions = list(module.list_versions())

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        module_id=self.id,
    )
    all_module_versions_info = self.list_pages_generator(
        self.STUB.ListModuleVersions,
        service_pb2.ListModuleVersionsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)

    for module_version_info in all_module_versions_info:
      module_version_info['id'] = module_version_info['module_version_id']
      del module_version_info['module_version_id']
      yield Module.from_auth_helper(
          self.auth_helper,
          module_id=self.id,
          **dict(self.kwargs, module_version=module_version_info))

  def __getattr__(self, name):
    return getattr(self.module_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.module_info, param)}" for param in init_params
        if hasattr(self.module_info, param)
    ]
    return f"Module Details: \n{', '.join(attribute_strings)}\n"



DATASET_UPLOAD_TASKS = [
    "visual_classification", "text_classification", "visual_detection", "visual_segmentation",
    "visual_captioning", "multimodal_dataset"
]

TASK_TO_ANNOTATION_TYPE = {
    "visual_classification": {
        "concepts": "labels"
    },
    "text_classification": {
        "concepts": "labels"
    },
    "visual_captioning": {
        "concepts": "labels"
    },
    "visual_detection": {
        "bboxes": "bboxes"
    },
    "visual_segmentation": {
        "polygons": "polygons"
    },
}

MAX_RETRIES = 2

CONTENT_TYPE = {"json": "application/json", "zip": "application/zip"}



import csv
import json
import os
import time
import uuid
from concurrent.futures import ThreadPoolExecutor, as_completed
from multiprocessing import cpu_count
from typing import Generator, List, Union

import requests
from clarifai_grpc.grpc.api import resources_pb2, service_pb2  # noqa: F401
from clarifai_grpc.grpc.api.resources_pb2 import Annotation, Audio, Image, Input, Text, Video
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2
from google.protobuf.json_format import MessageToDict
from google.protobuf.struct_pb2 import Struct
from requests.adapters import HTTPAdapter, Retry
from tqdm import tqdm

from clarifai.client.base import BaseClient
from clarifai.client.lister import Lister
from clarifai.constants.dataset import MAX_RETRIES
from clarifai.constants.input import MAX_UPLOAD_BATCH_SIZE
from clarifai.errors import UserError
from clarifai.utils.logging import logger
from clarifai.utils.misc import BackoffIterator, Chunker, clean_input_id


class Inputs(Lister, BaseClient):
  """Inputs is a class that provides access to Clarifai API endpoints related to Input information."""

  def __init__(self,
               user_id: str = None,
               app_id: str = None,
               logger_level: str = "INFO",
               base_url: str = "https://api.clarifai.com",
               pat: str = None,
               token: str = None,
               root_certificates_path: str = None,
               **kwargs):
    """Initializes an Input object.

    Args:
        user_id (str): A user ID for authentication.
        app_id (str): An app ID for the application to interact with.
        base_url (str): Base API url. Default "https://api.clarifai.com"
        pat (str): A personal access token for authentication. Can be set as env var CLARIFAI_PAT
        token (str): A session token for authentication. Accepts either a session token or a pat. Can be set as env var CLARIFAI_SESSION_TOKEN
        root_certificates_path (str): Path to the SSL root certificates file, used to establish secure gRPC connections.
        **kwargs: Additional keyword arguments to be passed to the Input
    """
    self.user_id = user_id
    self.app_id = app_id
    self.kwargs = {**kwargs}
    self.input_info = resources_pb2.Input(**self.kwargs)
    self.logger = logger
    BaseClient.__init__(
        self,
        user_id=self.user_id,
        app_id=self.app_id,
        base=base_url,
        pat=pat,
        token=token,
        root_certificates_path=root_certificates_path)
    Lister.__init__(self)

  @staticmethod
  def _get_proto(input_id: str,
                 dataset_id: str = None,
                 imagepb: Image = None,
                 video_pb: Video = None,
                 audio_pb: Audio = None,
                 text_pb: Text = None,
                 geo_info: List = None,
                 labels: List = None,
                 label_ids: List = None,
                 metadata: Struct = None) -> Input:
    """Create input proto for image data type.
        Args:
            input_id (str): The input ID for the input to create.
            dataset_id (str): The dataset ID for the dataset to add the input to.
            imagepb (Image): The image proto to be used for the input.
            video_pb (Video): The video proto to be used for the input.
            audio_pb (Audio): The audio proto to be used for the input.
            text_pb (Text): The text proto to be used for the input.
            geo_info (list): A list of longitude and latitude for the geo point.
            labels (list): A list of label names for the input.
            label_ids (list): A list of label ids for the input.
            metadata (Struct): A Struct of metadata for the input.
        Returns:
            Input: An Input object for the specified input ID.
        """
    assert geo_info is None or isinstance(
        geo_info, list), "geo_info must be a list of longitude and latitude"
    assert labels is None or isinstance(labels, list), "labels must be a list of strings"
    assert label_ids is None or isinstance(label_ids, list), "label_ids must be a list of strings"
    assert metadata is None or isinstance(metadata, Struct), "metadata must be a Struct"
    geo_pb = resources_pb2.Geo(geo_point=resources_pb2.GeoPoint(
        longitude=geo_info[0], latitude=geo_info[1])) if geo_info else None
    if labels:
      if not label_ids:
        concepts=[
            resources_pb2.Concept(
            id=_label, name=_label, value=1.)\
            for _label in labels
        ]
      else:
        assert len(labels) == len(label_ids), "labels and label_ids must be of the same length"
        concepts=[
            resources_pb2.Concept(
            id=label_id, name=_label, value=1.)\
            for label_id, _label in zip(label_ids, labels)
        ]
    else:
      concepts = None

    if dataset_id:
      return resources_pb2.Input(
          id=input_id,
          dataset_ids=[dataset_id],
          data=resources_pb2.Data(
              image=imagepb,
              video=video_pb,
              audio=audio_pb,
              text=text_pb,
              geo=geo_pb,
              concepts=concepts,
              metadata=metadata))

    return resources_pb2.Input(
        id=input_id,
        data=resources_pb2.Data(
            image=imagepb,
            video=video_pb,
            audio=audio_pb,
            text=text_pb,
            geo=geo_pb,
            concepts=concepts,
            metadata=metadata))

  @staticmethod
  def get_input_from_url(input_id: str,
                         image_url: str = None,
                         video_url: str = None,
                         audio_url: str = None,
                         text_url: str = None,
                         dataset_id: str = None,
                         **kwargs) -> Input:
    """Create input proto from url.

    Args:
        input_id (str): The input ID for the input to create.
        image_url (str): The url for the image.
        video_url (str): The url for the video.
        audio_url (str): The url for the audio.
        text_url (str): The url for the text.
            dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        Input: An Input object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_proto = Inputs.get_input_from_url(input_id = 'demo', image_url='https://samples.clarifai.com/metro-north.jpg')
    """
    if not any((image_url, video_url, audio_url, text_url)):
      raise ValueError(
          "At least one of image_url, video_url, audio_url, text_url must be provided.")
    image_pb = resources_pb2.Image(url=image_url) if image_url else None
    video_pb = resources_pb2.Video(url=video_url) if video_url else None
    audio_pb = resources_pb2.Audio(url=audio_url) if audio_url else None
    text_pb = resources_pb2.Text(url=text_url) if text_url else None
    return Inputs._get_proto(
        input_id=input_id,
        dataset_id=dataset_id,
        imagepb=image_pb,
        video_pb=video_pb,
        audio_pb=audio_pb,
        text_pb=text_pb,
        **kwargs)

  @staticmethod
  def get_input_from_file(input_id: str,
                          image_file: str = None,
                          video_file: str = None,
                          audio_file: str = None,
                          text_file: str = None,
                          dataset_id: str = None,
                          **kwargs) -> Input:
    """Create input proto from files.

    Args:
        input_id (str): The input ID for the input to create.
        image_file (str): The file_path for the image.
        video_file (str): The file_path for the video.
        audio_file (str): The file_path for the audio.
        text_file (str): The file_path for the text.
        dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        Input: An Input object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_proto = Inputs.get_input_from_file(input_id = 'demo', video_file='file_path')
    """
    if not any((image_file, video_file, audio_file, text_file)):
      raise ValueError(
          "At least one of image_file, video_file, audio_file, text_file must be provided.")
    image_pb = resources_pb2.Image(base64=open(image_file, 'rb').read()) if image_file else None
    video_pb = resources_pb2.Video(base64=open(video_file, 'rb').read()) if video_file else None
    audio_pb = resources_pb2.Audio(base64=open(audio_file, 'rb').read()) if audio_file else None
    text_pb = resources_pb2.Text(raw=open(text_file, 'rb').read()) if text_file else None
    return Inputs._get_proto(
        input_id=input_id,
        dataset_id=dataset_id,
        imagepb=image_pb,
        video_pb=video_pb,
        audio_pb=audio_pb,
        text_pb=text_pb,
        **kwargs)

  @staticmethod
  def get_input_from_bytes(input_id: str,
                           image_bytes: bytes = None,
                           video_bytes: bytes = None,
                           audio_bytes: bytes = None,
                           text_bytes: bytes = None,
                           dataset_id: str = None,
                           **kwargs) -> Input:
    """Create input proto from bytes.

    Args:
        input_id (str): The input ID for the input to create.
        image_bytes (str): The bytes for the image.
        video_bytes (str): The bytes for the video.
        audio_bytes (str): The bytes for the audio.
        text_bytes (str): The bytes for the text.
        dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        Input: An Input object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> image = open('demo.jpg', 'rb').read()
        >>> video = open('demo.mp4', 'rb').read()
        >>> input_proto = Inputs.get_input_from_bytes(input_id = 'demo',image_bytes =image, video_bytes=video)
    """
    if not any((image_bytes, video_bytes, audio_bytes, text_bytes)):
      raise ValueError(
          "At least one of image_bytes, video_bytes, audio_bytes, text_bytes must be provided.")
    image_pb = resources_pb2.Image(base64=image_bytes) if image_bytes else None
    video_pb = resources_pb2.Video(base64=video_bytes) if video_bytes else None
    audio_pb = resources_pb2.Audio(base64=audio_bytes) if audio_bytes else None
    text_pb = resources_pb2.Text(raw=text_bytes) if text_bytes else None
    return Inputs._get_proto(
        input_id=input_id,
        dataset_id=dataset_id,
        imagepb=image_pb,
        video_pb=video_pb,
        audio_pb=audio_pb,
        text_pb=text_pb,
        **kwargs)

  @staticmethod
  def get_image_inputs_from_folder(folder_path: str, dataset_id: str = None,
                                   labels: bool = False) -> List[Input]:  #image specific
    """Create input protos for image data type from folder.

    Args:
        folder_path (str): Path to the folder containing images.

    Returns:
        list of Input: A list of Input objects for the specified folder.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_protos = Inputs.get_image_inputs_from_folder(folder_path='demo_folder')
    """
    input_protos = []
    labels = [folder_path.split('/')[-1]] if labels else None
    for filename in os.listdir(folder_path):
      if filename.split('.')[-1] not in ['jpg', 'jpeg', 'png', 'tiff', 'webp']:
        continue
      input_id = clean_input_id(filename.split('.')[0])
      image_pb = resources_pb2.Image(base64=open(os.path.join(folder_path, filename), 'rb').read())
      input_protos.append(
          Inputs._get_proto(
              input_id=input_id, dataset_id=dataset_id, imagepb=image_pb, labels=labels))
    return input_protos

  @staticmethod
  def get_text_input(input_id: str, raw_text: str, dataset_id: str = None,
                     **kwargs) -> Text:  #text specific
    """Create input proto for text data type from raw text.

    Args:
        input_id (str): The input ID for the input to create.
        raw_text (str): The raw text input.
        dataset_id (str): The dataset ID for the dataset to add the input to.
        **kwargs: Additional keyword arguments to be passed to the Input

    Returns:
        Text: An Input object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_protos = Inputs.get_text_input(input_id = 'demo', raw_text = 'This is a test')
    """
    text_pb = resources_pb2.Text(raw=raw_text)
    return Inputs._get_proto(input_id=input_id, dataset_id=dataset_id, text_pb=text_pb, **kwargs)

  @staticmethod
  def get_multimodal_input(input_id: str,
                           raw_text: str = None,
                           text_bytes: bytes = None,
                           image_url: str = None,
                           image_bytes: bytes = None,
                           dataset_id: str = None,
                           **kwargs) -> Text:
    """Create input proto for text and image from bytes or url.

    Args:
        input_id (str): The input ID for the input to create.
        raw_text (str): The raw text input.
        text_bytes (str): The bytes for the text.
        image_url (str): The url for the image.
        image_bytes (str): The bytes for the image.
        dataset_id (str): The dataset ID for the dataset to add the input to.
        **kwargs: Additional keyword arguments to be passed to the Input

    Returns:
        Input: An Input object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_protos = Inputs.get_multimodal_input(input_id = 'demo', raw_text = 'What time of day is it?', image_url='https://samples.clarifai.com/metro-north.jpg')
    """
    if (image_bytes and image_url) or (not image_bytes and not image_url):
      return UserError("Please supply only one of image_bytes or image_url, and not both.")
    if (text_bytes and raw_text) or (not text_bytes and not raw_text):
      return UserError("Please supply only one of text_bytes or raw_text, and not both.")

    image_pb = resources_pb2.Image(base64=image_bytes) if image_bytes else resources_pb2.Image(
        url=image_url) if image_url else None
    text_pb = resources_pb2.Text(raw=text_bytes) if text_bytes else resources_pb2.Text(
        raw=raw_text) if raw_text else None
    return Inputs._get_proto(
        input_id=input_id, dataset_id=dataset_id, imagepb=image_pb, text_pb=text_pb, **kwargs)

  @staticmethod
  def get_inputs_from_csv(csv_path: str,
                          input_type: str = 'text',
                          csv_type: str = 'raw',
                          dataset_id: str = None,
                          labels: str = True) -> List[Text]:
    """Create input protos from csv.

    Args:
        csv_path (str): Path to the csv file.
        input_type (str): Type of input. Options: 'text', 'image', 'video', 'audio'.
        csv_type (str): Type of csv file. Options: 'raw', 'url', 'file_path'.
        dataset_id (str): The dataset ID for the dataset to add the input to.
        labels (str): True if csv file has labels column.

    Returns:
        inputs: List of inputs

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_protos = Inputs.get_inputs_from_csv(csv_path='filepath', input_type='text', csv_type='raw')
    """
    input_protos = []
    with open(csv_path) as _file:
      reader = csv.DictReader(_file, delimiter=',', quotechar='"')
      columns = reader.fieldnames
      for column in columns:
        if column not in ['inputid', 'input', 'concepts', 'metadata', 'geopoints']:
          raise UserError(
              "CSV file may have 'inputid', 'input', 'concepts', 'metadata', 'geopoints' columns. Does not support '{}' column".
              format(column))
      for id, input in enumerate(reader):
        if labels:
          labels_list = input['concepts'].split(',')
          labels = labels_list if len(input['concepts']) > 0 else None
        else:
          labels = None

        if 'metadata' in columns:
          if len(input['metadata']) > 0:
            metadata_str = input['metadata'].replace("'", '"')
            try:
              metadata_dict = json.loads(metadata_str)
            except json.decoder.JSONDecodeError:
              raise UserError("metadata column in CSV file should be a valid json")
            metadata = Struct()
            metadata.update(metadata_dict)
          else:
            metadata = None
        else:
          metadata = None

        if 'geopoints' in columns:
          if len(input['geopoints']) > 0:
            geo_points = input['geopoints'].split(',')
            geo_points = [float(geo_point) for geo_point in geo_points]
            geo_info = geo_points if len(geo_points) == 2 else UserError(
                "geopoints column in CSV file should have longitude,latitude")
          else:
            geo_info = None
        else:
          geo_info = None

        input_id = input['inputid'] if 'inputid' in columns else uuid.uuid4().hex
        text = input['input'] if input_type == 'text' else None
        image = input['input'] if input_type == 'image' else None
        video = input['input'] if input_type == 'video' else None
        audio = input['input'] if input_type == 'audio' else None

        if csv_type == 'raw':
          input_protos.append(
              Inputs.get_text_input(
                  input_id=input_id,
                  raw_text=text,
                  dataset_id=dataset_id,
                  labels=labels,
                  metadata=metadata,
                  geo_info=geo_info))
        elif csv_type == 'url':
          input_protos.append(
              Inputs.get_input_from_url(
                  input_id=input_id,
                  image_url=image,
                  text_url=text,
                  audio_url=audio,
                  video_url=video,
                  dataset_id=dataset_id,
                  labels=labels,
                  metadata=metadata,
                  geo_info=geo_info))
        else:
          input_protos.append(
              Inputs.get_input_from_file(
                  input_id=input_id,
                  image_file=image,
                  text_file=text,
                  audio_file=audio,
                  video_file=video,
                  dataset_id=dataset_id,
                  labels=labels,
                  metadata=metadata,
                  geo_info=geo_info))

    return input_protos

  @staticmethod
  def get_text_inputs_from_folder(folder_path: str, dataset_id: str = None,
                                  labels: bool = False) -> List[Text]:  #text specific
    """Create input protos for text data type from folder.

    Args:
        folder_path (str): Path to the folder containing text.

    Returns:
        list of Input: A list of Input objects for the specified folder.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_protos = Inputs.get_text_inputs_from_folder(folder_path='demo_folder')
    """
    input_protos = []
    labels = [folder_path.split('/')[-1]] if labels else None
    for filename in os.listdir(folder_path):
      if filename.split('.')[-1] != 'txt':
        continue
      input_id = clean_input_id(filename.split('.')[0])
      text_pb = resources_pb2.Text(raw=open(os.path.join(folder_path, filename), 'rb').read())
      input_protos.append(
          Inputs._get_proto(
              input_id=input_id, dataset_id=dataset_id, text_pb=text_pb, labels=labels))
    return input_protos

  @staticmethod
  def get_bbox_proto(input_id: str,
                     label: str,
                     bbox: List,
                     label_id: str = None,
                     annot_id: str = None) -> Annotation:
    """Create an annotation proto for each bounding box, label input pair.

    Args:
        input_id (str): The input ID for the annotation to create.
        label (str): annotation label name
        bbox (List): a list of a single bbox's coordinates. # bbox ordering: [xmin, ymin, xmax, ymax]
        label_id (str): annotation label ID
        annot_id (str): annotation ID

    Returns:
        An annotation object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> Inputs.get_bbox_proto(input_id='demo', label='demo', bbox=[x_min, y_min, x_max, y_max])
    """
    if not isinstance(bbox, list):
      raise UserError("must be a list of bbox cooridnates")
    annot_data = resources_pb2.Data(regions=[
        resources_pb2.Region(
            region_info=resources_pb2.RegionInfo(bounding_box=resources_pb2.BoundingBox(
                # bbox ordering: [xmin, ymin, xmax, ymax]
                # top_row must be less than bottom row
                # left_col must be less than right col
                top_row=bbox[1],  #y_min
                left_col=bbox[0],  #x_min
                bottom_row=bbox[3],  #y_max
                right_col=bbox[2]  #x_max
            )),
            data=resources_pb2.Data(concepts=[
                resources_pb2.Concept(id=label, name=label, value=1.)
                if not label_id else resources_pb2.Concept(id=label_id, name=label, value=1.)
            ]))
    ])
    if annot_id:
      input_annot_proto = resources_pb2.Annotation(id=annot_id, input_id=input_id, data=annot_data)
    else:
      input_annot_proto = resources_pb2.Annotation(input_id=input_id, data=annot_data)

    return input_annot_proto

  @staticmethod
  def get_mask_proto(input_id: str,
                     label: str,
                     polygons: List[List[float]],
                     label_id: str = None,
                     annot_id: str = None) -> Annotation:
    """Create an annotation proto for each polygon box, label input pair.

    Args:
        input_id (str): The input ID for the annotation to create.
        label (str): annotation label name
        polygons (List): Polygon x,y points iterable
        label_id (str): annotation label ID
        annot_id (str): annotation ID

    Returns:
        An annotation object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> Inputs.get_mask_proto(input_id='demo', label='demo', polygons=[[[x,y],...,[x,y]],...])
    """
    if not isinstance(polygons, list):
      raise UserError("polygons must be a list of points")
    annot_data = resources_pb2.Data(regions=[
        resources_pb2.Region(
            region_info=resources_pb2.RegionInfo(polygon=resources_pb2.Polygon(
                points=[
                    resources_pb2.Point(
                        row=_point[1],  # row is y point
                        col=_point[0],  # col is x point
                        visibility="VISIBLE") for _point in polygons
                ])),
            data=resources_pb2.Data(concepts=[
                resources_pb2.Concept(id=label, name=label, value=1.)
                if not label_id else resources_pb2.Concept(id=label_id, name=label, value=1.)
            ]))
    ])
    if annot_id:
      input_mask_proto = resources_pb2.Annotation(id=annot_id, input_id=input_id, data=annot_data)
    else:
      input_mask_proto = resources_pb2.Annotation(input_id=input_id, data=annot_data)

    return input_mask_proto

  def get_input(self, input_id: str) -> Input:
    """Get Input object of input with input_id provided from the app.

    Args:
        input_id (str): The input ID for the annotation to get.

    Returns:
        Input: An Input object for the specified input ID.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_obj = Inputs(user_id = 'user_id', app_id = 'demo_app')
        >>> input_obj.get_input(input_id='demo')
    """
    request = service_pb2.GetInputRequest(user_app_id=self.user_app_id, input_id=input_id)
    response = self._grpc_request(self.STUB.GetInput, request)
    return response.input

  def upload_from_url(self,
                      input_id: str,
                      image_url: str = None,
                      video_url: str = None,
                      audio_url: str = None,
                      text_url: str = None,
                      dataset_id: str = None,
                      **kwargs) -> str:
    """Upload input from url.

    Args:
        input_id (str): The input ID for the input to create.
        image_url (str): The url for the image.
        video_url (str): The url for the video.
        audio_url (str): The url for the audio.
        text_url (str): The url for the text.
        dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        input_job_id: job id for the upload request.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_obj = Inputs(user_id = 'user_id', app_id = 'demo_app')
        >>> input_obj.upload_from_url(input_id='demo', image_url='https://samples.clarifai.com/metro-north.jpg')
    """
    input_pb = self.get_input_from_url(input_id, image_url, video_url, audio_url, text_url,
                                       dataset_id, **kwargs)
    return self.upload_inputs([input_pb])

  def upload_from_file(self,
                       input_id: str,
                       image_file: str = None,
                       video_file: str = None,
                       audio_file: str = None,
                       text_file: str = None,
                       dataset_id: str = None,
                       **kwargs) -> str:
    """Upload input from file.

    Args:
        input_id (str): The input ID for the input to create.
        image_file (str): The file for the image.
        video_file (str): The file for the video.
        audio_file (str): The file for the audio.
        text_file (str): The file for the text.
        dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        input_job_id: job id for the upload request.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_obj = Inputs(user_id = 'user_id', app_id = 'demo_app')
        >>> input_obj.upload_from_file(input_id='demo', audio_file='demo.mp3')
    """
    input_pb = self.get_input_from_file(input_id, image_file, video_file, audio_file, text_file,
                                        dataset_id, **kwargs)
    return self.upload_inputs([input_pb])

  def upload_from_bytes(self,
                        input_id: str,
                        image_bytes: bytes = None,
                        video_bytes: bytes = None,
                        audio_bytes: bytes = None,
                        text_bytes: bytes = None,
                        dataset_id: str = None,
                        **kwargs) -> str:
    """Upload input from bytes.

    Args:
        input_id (str): The input ID for the input to create.
        image_bytes (str): The bytes for the image.
        video_bytes (str): The bytes for the video.
        audio_bytes (str): The bytes for the audio.
        text_bytes (str): The bytes for the text.
        dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        input_job_id: job id for the upload request.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_obj = Inputs(user_id = 'user_id', app_id = 'demo_app')
        >>> image = open('demo.jpg', 'rb').read()
        >>> input_obj.upload_from_bytes(input_id='demo', image_bytes=image)
    """
    input_pb = self.get_input_from_bytes(input_id, image_bytes, video_bytes, audio_bytes,
                                         text_bytes, dataset_id, **kwargs)
    return self.upload_inputs([input_pb])

  def upload_text(self, input_id: str, raw_text: str, dataset_id: str = None,
                  **kwargs) -> str:  #text specific
    """Upload text from raw text.

    Args:
        input_id (str): The input ID for the input to create.
        raw_text (str): The raw text.
        dataset_id (str): The dataset ID for the dataset to add the input to.

    Returns:
        input_job_id (str): job id for the upload request.

    Example:
        >>> from clarifai.client.input import Inputs
        >>> input_obj = Inputs(user_id = 'user_id', app_id = 'demo_app')
        >>> input_obj.upload_text(input_id = 'demo', raw_text = 'This is a test')
    """
    input_pb = self._get_proto(
        input_id=input_id,
        dataset_id=dataset_id,
        text_pb=resources_pb2.Text(raw=raw_text),
        **kwargs)
    return self.upload_inputs([input_pb])

  def upload_inputs(self, inputs: List[Input], show_log: bool = True) -> str:
    """Upload list of input objects to the app.

    Args:
        inputs (list): List of input objects to upload.
        show_log (bool): Show upload status log.

    Returns:
        input_job_id: job id for the upload request.
    """
    if not isinstance(inputs, list):
      raise UserError("inputs must be a list of Input objects")
    if len(inputs) > MAX_UPLOAD_BATCH_SIZE:
      raise UserError(
          f"Number of inputs to upload exceeds the maximum batch size of {MAX_UPLOAD_BATCH_SIZE}. Please reduce batch size."
      )
    input_job_id = uuid.uuid4().hex  # generate a unique id for this job
    request = service_pb2.PostInputsRequest(
        user_app_id=self.user_app_id, inputs=inputs, inputs_add_job_id=input_job_id)
    response = self._grpc_request(self.STUB.PostInputs, request)
    if response.status.code != status_code_pb2.SUCCESS:
      if show_log:
        self.logger.warning(response)
      else:
        return input_job_id, response
    else:
      if show_log:
        self.logger.info("\nInputs Uploaded\n%s", response.status)

    return input_job_id, response

  def patch_inputs(self, inputs: List[Input], action: str = 'merge') -> None:
    """Patch list of input objects to the app.

    Args:
        inputs (list): List of input objects to upload.
        action (str): Action to perform on the input. Options: 'merge', 'overwrite', 'remove'.

    Returns:
        response: Response from the grpc request.
    """
    if not isinstance(inputs, list):
      raise UserError("inputs must be a list of Input objects")
    request = service_pb2.PatchInputsRequest(
        user_app_id=self.user_app_id, inputs=inputs, action=action)
    response = self._grpc_request(self.STUB.PatchInputs, request)
    if response.status.code != status_code_pb2.SUCCESS:
      try:
        self.logger.warning(f"Patch inputs failed, status: {response.annotations[0].status}")
      except Exception:
        self.logger.warning(f"Patch inputs failed, status: {response.status}")
    else:
      self.logger.info("\nPatch Inputs Successful\n%s", response.status)

  def upload_annotations(self, batch_annot: List[resources_pb2.Annotation], show_log: bool = True
                        ) -> Union[List[resources_pb2.Annotation], List[None]]:
    """Upload image annotations to app.

    Args:
        batch_annot: annot batch protos

    Returns:
        retry_upload: failed annot upload
    """
    retry_upload = []  # those that fail to upload are stored for retries
    request = service_pb2.PostAnnotationsRequest(
        user_app_id=self.user_app_id, annotations=batch_annot)
    response = self._grpc_request(self.STUB.PostAnnotations, request)
    response_dict = MessageToDict(response)
    if response.status.code != status_code_pb2.SUCCESS:
      try:
        for annot in response_dict["annotations"]:
          if annot['status']['code'] != status_code_pb2.ANNOTATION_SUCCESS:
            self.logger.warning(f"Post annotations failed, status: {annot['status']}")
      except Exception:
        self.logger.warning(f"Post annotations failed due to {response.status}")
      finally:
        retry_upload.extend(batch_annot)
    else:
      if show_log:
        self.logger.info("\nAnnotations Uploaded\n%s", response.status)

    return retry_upload

  def patch_annotations(self, batch_annot: List[resources_pb2.Annotation],
                        action: str = 'merge') -> None:
    """Patch image annotations to app.

    Args:
        batch_annot: annot batch protos
        action (str): Action to perform on the input. Options: 'merge', 'overwrite', 'remove'.

    """
    if not isinstance(batch_annot, list):
      raise UserError("batch_annot must be a list of Annotation objects")
    request = service_pb2.PatchAnnotationsRequest(
        user_app_id=self.user_app_id, annotations=batch_annot, action=action)
    response = self._grpc_request(self.STUB.PatchAnnotations, request)
    response_dict = MessageToDict(response)
    if response.status.code != status_code_pb2.SUCCESS:
      try:
        for annot in response_dict["annotations"]:
          if annot['status']['code'] != status_code_pb2.ANNOTATION_SUCCESS:
            self.logger.warning(f"Patch annotations failed, status: {annot['status']}")
      except Exception:
        self.logger.warning(f"Patch annotations failed due to {response.status}")
    else:
      self.logger.info("\nPatch Annotations Uploaded Successful\n%s", response.status)

  def patch_concepts(self,
                     concept_ids: List[str],
                     labels: List[str] = [],
                     values: List[float] = [],
                     action: str = 'overwrite') -> None:
    """Patch concepts to app.

    Args:
        concept_ids:  A list of concept
        labels: A list of label names
        values: concept value
        action (str): Action to perform on the input. Options: 'overwrite'.

    """
    if not labels:
      labels = list(concept_ids)
    if values:
      concepts=[
              resources_pb2.Concept(
              id=concept_id, name=label, value=value)\
              for concept_id, label, value in zip(concept_ids, labels, values)
          ]
    else:
      concepts=[
              resources_pb2.Concept(
              id=concept_id, name=label, value=1.)\
              for concept_id, label in zip(concept_ids, labels)
          ]
    request = service_pb2.PatchConceptsRequest(
        user_app_id=self.user_app_id, concepts=concepts, action=action)
    response = self._grpc_request(self.STUB.PatchConcepts, request)
    if response.status.code != status_code_pb2.SUCCESS:
      self.logger.warning(f"Patch Concepts failed, status: {response.status.details}")
    else:
      self.logger.info("\nPatch Concepts Successful\n%s", response.status)

  def _upload_batch(self, inputs: List[Input]) -> List[Input]:
    """Upload a batch of input objects to the app.

    Args:
        inputs (List[Input]): List of input objects to upload.

    Returns:
        input_job_id: job id for the upload request.
    """
    input_job_id, _ = self.upload_inputs(inputs, False)
    self._wait_for_inputs(input_job_id)
    failed_inputs = self._delete_failed_inputs(inputs)

    return failed_inputs

  def delete_inputs(self, inputs: List[Input]) -> None:
    """Delete list of input objects from the app.

    Args:
        input_ids (Input): List of input objects to delete.

    Example:
        >>> from clarifai.client.user import User
        >>> input_obj = User(user_id="user_id").app(app_id="app_id").inputs()
        >>> input_obj.delete_inputs(list(input_obj.list_inputs()))
    """
    if not isinstance(inputs, list):
      raise UserError("input_ids must be a list of input ids")
    inputs_ids = [input.id for input in inputs]
    request = service_pb2.DeleteInputsRequest(user_app_id=self.user_app_id, ids=inputs_ids)
    response = self._grpc_request(self.STUB.DeleteInputs, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nInputs Deleted\n%s", response.status)

  def delete_annotations(self, input_ids: List[str], annotation_ids: List[str] = []) -> None:
    """Delete list of annotations of input objects from the app.

    Args:
        input_ids (Input): List of input objects for which annotations to delete.
        annotation_ids (List[str]): List of annotation ids to delete.

    Example:
        >>> from clarifai.client.user import User
        >>> input_obj = User(user_id="user_id").app(app_id="app_id").inputs()
        >>> input_obj.delete_annotations(input_ids=['input_id_1', 'input_id_2'])

    Note:
        'annotation_ids' are optional but if the are provided, the number and order in
        'annotation_ids' and 'input_ids' should match
    """
    if not isinstance(input_ids, list):
      raise UserError("input_ids must be a list of input ids")
    if annotation_ids and len(input_ids) != len(annotation_ids):
      raise UserError("Number of provided annotation_ids and input_ids should match.")
    request = service_pb2.DeleteAnnotationsRequest(
        user_app_id=self.user_app_id, ids=annotation_ids, input_ids=input_ids)
    response = self._grpc_request(self.STUB.DeleteAnnotations, request)
    if response.status.code != status_code_pb2.SUCCESS:
      raise Exception(response.status)
    self.logger.info("\nAnnotations Deleted\n%s", response.status)

  def download_inputs(self, inputs: List[Input]) -> List[bytes]:
    """Download list of input objects from the app.

    Args:
        input_ids (Input): List of input objects to download.

    Example:
        >>> from clarifai.client.user import User
        >>> input_obj = User(user_id="user_id").app(app_id="app_id").inputs()
        >>> input_obj.download_inputs(list(input_obj.list_inputs()))
    """
    if not isinstance(inputs, list):
      raise UserError("input_ids must be a list of input ids")
    final_inputs = []
    #initiate session
    session = requests.Session()
    retries = Retry(total=3, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
    session.mount('https://', HTTPAdapter(max_retries=retries))
    session.headers.update({'Authorization': self.metadata[0][1]})
    # download inputs
    data_types = ['image', 'video', 'audio', 'text']
    for input in inputs:
      for data_type in data_types:
        url = getattr(input.data, data_type).url
        if url:
          response = session.get(url, stream=True)
          if response.status_code == 200:
            final_inputs.append(response.content)

    return final_inputs

  def list_inputs(self,
                  dataset_id: str = None,
                  page_no: int = None,
                  per_page: int = None,
                  input_type: str = None) -> Generator[Input, None, None]:
    """Lists all the inputs for the app.

    Args:
        dataset_id (str): The dataset ID for the dataset to list inputs from.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.
        input_type (str): The type of input to list. Options: 'image', 'video', 'audio', 'text'.

    Yields:
        Input: Input objects for the app.

    Example:
        >>> from clarifai.client.user import User
        >>> input_obj = User(user_id="user_id").app(app_id="app_id").inputs()
        >>> all_inputs = list(input_obj.list_inputs(input_type='image'))

    Note:
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    if input_type not in ['image', 'text', 'video', 'audio', None]:
      raise UserError('Invalid input type, it should be image,text,audio or video')
    if dataset_id:
      request_data = dict(user_app_id=self.user_app_id, dataset_id=dataset_id)
      all_inputs_info = self.list_pages_generator(
          self.STUB.ListDatasetInputs,
          service_pb2.ListDatasetInputsRequest,
          request_data,
          per_page=per_page,
          page_no=page_no)
    else:
      request_data = dict(user_app_id=self.user_app_id)
      all_inputs_info = self.list_pages_generator(
          self.STUB.ListInputs,
          service_pb2.ListInputsRequest,
          request_data,
          per_page=per_page,
          page_no=page_no)
    for input_info in all_inputs_info:
      input_info['id'] = input_info.pop('dataset_input_id') if dataset_id else input_info.pop(
          'input_id')
      if input_type:
        if input_type not in input_info['data'].keys():
          continue
      yield resources_pb2.Input(**input_info)

  def list_annotations(self,
                       batch_input: List[Input] = None,
                       page_no: int = None,
                       per_page: int = None) -> Generator[Annotation, None, None]:
    """Lists all the annotations for the app.

    Args:
        batch_input (List[Input]): The input objects to list annotations from.
        page_no (int): The page number to list.
        per_page (int): The number of items per page.

    Yields:
        Annotation: Annotation objects for the app.

    Example:
        >>> from clarifai.client.user import User
        >>> input_obj = User(user_id="user_id").app(app_id="app_id").inputs()
        >>> all_inputs = list(input_obj.list_inputs(input_type='image'))
        >>> all_annotations = list(input_obj.list_annotations(batch_input=all_inputs))

    Note:
        If batch_input is not given, then lists all the annotations for the app.
        Defaults to 16 per page if page_no is specified and per_page is not specified.
        If both page_no and per_page are None, then lists all the resources.
    """
    request_data = dict(
        user_app_id=self.user_app_id,
        input_ids=[input.id for input in batch_input] if batch_input else None)
    all_annotations_info = self.list_pages_generator(
        self.STUB.ListAnnotations,
        service_pb2.ListAnnotationsRequest,
        request_data,
        per_page=per_page,
        page_no=page_no)
    for annotations_info in all_annotations_info:
      annotations_info['id'] = annotations_info.pop('annotation_id')
      yield Annotation(**annotations_info)

  def _bulk_upload(self, inputs: List[Input], batch_size: int = 128) -> None:
    """Uploads process for large number of inputs.

    Args:
        inputs (List[Input]): input protos
        batch_size (int): batch size for each request
    """
    num_workers: int = min(10, cpu_count())  # limit max workers to 10
    batch_size = min(128, batch_size)  # limit max protos in a req
    chunked_inputs = Chunker(inputs, batch_size).chunk()
    with ThreadPoolExecutor(max_workers=num_workers) as executor:
      with tqdm(total=len(chunked_inputs), desc='Uploading inputs') as progress:
        # Submit all jobs to the executor and store the returned futures
        futures = [
            executor.submit(self._upload_batch, batch_input_ids)
            for batch_input_ids in chunked_inputs
        ]

        for job in as_completed(futures):
          retry_input_proto = job.result()
          self._retry_uploads(retry_input_proto)
          progress.update()

  def _wait_for_inputs(self, input_job_id: str) -> bool:
    """Wait for inputs to be processed. Cancel Job if timeout > 30 minutes.

    Args:
        input_job_id (str): Upload Input Job ID

    Returns:
        True if inputs are processed, False otherwise
    """
    backoff_iterator = BackoffIterator(10)
    max_retries = 10
    start_time = time.time()
    while True:
      request = service_pb2.GetInputsAddJobRequest(user_app_id=self.user_app_id, id=input_job_id)
      response = self._grpc_request(self.STUB.GetInputsAddJob, request)

      if time.time() - start_time > 60 * 30 or max_retries == 0:  # 30 minutes timeout
        self._grpc_request(self.STUB.CancelInputsAddJob,
                           service_pb2.CancelInputsAddJobRequest(
                               user_app_id=self.user_app_id, id=input_job_id))  #Cancel Job
        return False
      if response.status.code != status_code_pb2.SUCCESS:
        max_retries -= 1
        self.logger.warning(f"Get input job failed, status: {response.status.details}\n")
        continue
      if response.inputs_add_job.progress.in_progress_count == 0 and response.inputs_add_job.progress.pending_count == 0:
        return True
      else:
        time.sleep(next(backoff_iterator))

  def _retry_uploads(self, failed_inputs: List[Input]) -> None:
    """Retry failed uploads.

    Args:
        failed_inputs (List[Input]): failed input protos
    """
    for _retry in range(MAX_RETRIES):
      if failed_inputs:
        self.logger.info(f"Retrying upload for {len(failed_inputs)} Failed inputs..\n")
        failed_inputs = self._upload_batch(failed_inputs)

    self.logger.warning(f"Failed to upload {len(failed_inputs)} inputs..\n ")

  def _delete_failed_inputs(self, inputs: List[Input]) -> List[Input]:
    """Delete failed input ids from clarifai platform dataset.

    Args:
        inputs (List[Input]): batch input protos

    Returns:
        failed_inputs: failed inputs
    """
    input_ids = [input.id for input in inputs]
    success_status = status_pb2.Status(code=status_code_pb2.INPUT_DOWNLOAD_SUCCESS)
    request = service_pb2.ListInputsRequest(
        ids=input_ids,
        per_page=len(input_ids),
        user_app_id=self.user_app_id,
        status=success_status)
    response = self._grpc_request(self.STUB.ListInputs, request)
    response_dict = MessageToDict(response)
    success_inputs = response_dict.get('inputs', [])

    success_input_ids = [input.get('id') for input in success_inputs]
    failed_inputs = [input for input in inputs if input.id not in success_input_ids]
    #delete failed inputs
    self._grpc_request(self.STUB.DeleteInputs,
                       service_pb2.DeleteInputsRequest(
                           user_app_id=self.user_app_id, ids=[input.id
                                                              for input in failed_inputs]))

    return failed_inputs

  def __getattr__(self, name):
    return getattr(self.input_info, name)

  def __str__(self):
    init_params = [param for param in self.kwargs.keys()]
    attribute_strings = [
        f"{param}={getattr(self.input_info, param)}" for param in init_params
        if hasattr(self.input_info, param)
    ]
    return f"Input Details: \n{', '.join(attribute_strings)}\n"



import json
import os
import sys

import click
import yaml

from clarifai.utils.cli import AliasedGroup, TableFormatter, load_command_modules
from clarifai.utils.config import Config, Context
from clarifai.utils.constants import DEFAULT_CONFIG
from clarifai.utils.logging import logger


#@click.group(cls=CustomMultiGroup)
@click.group(cls=AliasedGroup)
@click.option('--config', default=DEFAULT_CONFIG)
@click.pass_context
def cli(ctx, config):
  """Clarifai CLI"""
  ctx.ensure_object(dict)
  if os.path.exists(config):
    cfg = Config.from_yaml(filename=config)
    ctx.obj = cfg
  else:
    cfg = Config(
        filename=config,
        current_context='default',
        contexts={
            'default':
                Context(
                    'default',
                    CLARIFAI_PAT=os.environ.get('CLARIFAI_PAT', ''),
                    CLARIFAI_USER_ID=os.environ.get('CLARIFAI_USER_ID', ''),
                    CLARIFAI_API_BASE=os.environ.get('CLARIFAI_API_BASE', 'api.clarifai.com'),
                )
        })
    try:
      cfg.to_yaml(config)
    except Exception:
      logger.warning("Could not write configuration to disk. Could be a read only file system.")
    ctx.obj = cfg  # still have the default config even if couldn't write.


@cli.command()
@click.argument('shell', type=click.Choice(['bash', 'zsh']))
def shell_completion(shell):
  """Shell completion script"""
  os.system(f"_CLARIFAI_COMPLETE={shell}_source clarifai")


@cli.group(['cfg'], cls=AliasedGroup)
def config():
  """Manage CLI configuration"""


@config.command(['e'])
@click.pass_context
def edit(ctx):
  """Edit the configuration file"""
  os.system(f'{os.environ.get("EDITOR", "vi")} {ctx.obj.filename}')


@config.command(['current'])
@click.option('-o', '--output-format', default='name', type=click.Choice(['name', 'json', 'yaml']))
@click.pass_context
def current_context(ctx, output_format):
  """Get the current context"""
  if output_format == 'name':
    print(ctx.obj.current_context)
  else:
    if output_format == 'json':
      print(json.dumps(ctx.obj.contexts[ctx.obj.current_context].to_serializable_dict()))
    else:
      print(yaml.safe_dump(ctx.obj.contexts[ctx.obj.current_context].to_serializable_dict()))


@config.command(['list', 'ls'])
@click.option(
    '-o', '--output-format', default='wide', type=click.Choice(['wide', 'name', 'json', 'yaml']))
@click.pass_context
def get_contexts(ctx, output_format):
  """Get all contexts"""
  if output_format == 'wide':
    formatter = TableFormatter(custom_columns={
        '': lambda c: '*' if c.name == ctx.obj.current_context else '',
        'NAME': lambda c: c.name,
        'USER_ID': lambda c: c.user_id,
        'API_BASE': lambda c: c.api_base,
        'PAT': lambda c: pat_display(c.pat),
    })
    print(formatter.format(ctx.obj.contexts.values(), fmt="plain"))
  elif output_format == 'name':
    print('\n'.join(ctx.obj.contexts))
  elif output_format in ('json', 'yaml'):
    dicts = [v.__dict__ for c, v in ctx.obj.contexts.items()]
    for d in dicts:
      d.pop('pat')
    if output_format == 'json':
      print(json.dumps(dicts))
    elif output_format == 'yaml':
      print(yaml.safe_dump(dicts))


@config.command(['use'])
@click.argument('context-name', type=str)
@click.pass_context
def use_context(ctx, context_name):
  """Set the current context"""
  if context_name not in ctx.obj.contexts:
    raise click.UsageError('Context not found')
  ctx.obj.current_context = context_name
  ctx.obj.to_yaml()
  print(f'Set {context_name} as the current context')


@config.command(['cat'])
@click.option('-o', '--output-format', default='yaml', type=click.Choice(['yaml', 'json']))
@click.pass_obj
def dump(ctx_obj, output_format):
  """Dump the configuration to stdout"""
  if output_format == 'yaml':
    yaml.safe_dump(ctx_obj.to_dict(), sys.stdout)
  else:
    json.dump(ctx_obj.to_dict(), sys.stdout, indent=2)


@cli.command()
@click.argument('api_url', default="https://api.clarifai.com")
@click.option('--user_id', required=False, help='User ID')
@click.pass_context
def login(ctx, api_url, user_id):
  """Login command to set PAT and other configurations."""

  name = input('context name (default: "default"): ')
  user_id = user_id if user_id is not None else input('user id: ')
  pat = input_or_default(
      'personal access token value (default: "ENVVAR" to get our of env var rather than config): ',
      'ENVVAR')

  context = Context(
      name,
      CLARIFAI_API_BASE=api_url,
      CLARIFAI_USER_ID=user_id,
      CLARIFAI_PAT=pat,
  )

  if context.name == '':
    context.name = 'default'

  ctx.obj.contexts[context.name] = context
  ctx.obj.current_context = context.name

  ctx.obj.to_yaml()


@cli.group(cls=AliasedGroup)
def context():
  """Manage contexts"""


def pat_display(pat):
  return pat[:5] + "****"


@context.command(['ls'])
@click.pass_context
def list(ctx):
  """List available contexts"""
  formatter = TableFormatter(custom_columns={
      '': lambda c: '*' if c.name == ctx.obj.current_context else '',
      'NAME': lambda c: c.name,
      'USER_ID': lambda c: c.user_id,
      'API_BASE': lambda c: c.api_base,
      'PAT': lambda c: pat_display(c.pat)
  })
  print(formatter.format(ctx.obj.contexts.values(), fmt="plain"))


def input_or_default(prompt, default):
  value = input(prompt)
  return value if value else default


@context.command()
@click.argument('name')
@click.option('--user-id', required=False, help='User ID')
@click.option('--base-url', required=False, help='Base URL')
@click.option('--pat', required=False, help='Personal access token')
@click.pass_context
def create(
    ctx,
    name,
    user_id=None,
    base_url=None,
    pat=None,
):
  """Create a new context"""
  if name in ctx.obj.contexts:
    print(f'{name} already exists')
    exit(1)
  if not user_id:
    user_id = input('user id: ')
  if not base_url:
    base_url = input_or_default('base url (default: https://api.clarifai.com): ',
                                'https://api.clarifai.com')
  if not pat:
    pat = input_or_default(
        'personal access token value (default: "ENVVAR" to get our of env var rather than config): ',
        'ENVVAR')

  context = Context(name, CLARIFAI_USER_ID=user_id, CLARIFAI_API_BASE=base_url, CLARIFAI_PAT=pat)
  ctx.obj.contexts[context.name] = context
  ctx.obj.to_yaml()


# write a click command to delete a context
@context.command(['rm'])
@click.argument('name')
@click.pass_context
def delete(ctx, name):
  """Delete a context"""
  if name not in ctx.obj.contexts:
    print(f'{name} is not a valid context')
    exit(1)
  ctx.obj.contexts.pop(name)
  ctx.obj.to_yaml()
  print(f'{name} deleted')


@context.command()
@click.argument('name', type=str)
@click.pass_context
def use(ctx, name):
  """Set the current context"""
  if name not in ctx.obj.contexts:
    raise click.UsageError('Context not found')
  ctx.obj.current_context = name
  ctx.obj.to_yaml()
  print(f'Set {name} as the current context')


@cli.command()
@click.argument('script', type=str)
@click.option('--context', type=str, help='Context to use')
@click.pass_context
def run(ctx, script, context=None):
  """Execute a script with the current context's environment"""
  context = ctx.obj.current if not context else context
  cmd = f'CLARIFAI_USER_ID={context.user_id} CLARIFAI_API_BASE={context.base_url} CLARIFAI_PAT={context.pat} '
  cmd += ' '.join([f'{k}={v}' for k, v in context.env.items()])
  cmd += f' {script}'
  os.system(cmd)


# Import the CLI commands to register them
load_command_modules()


def main():
  cli()



import click

from clarifai.cli.base import cli
from clarifai.client.compute_cluster import ComputeCluster
from clarifai.client.nodepool import Nodepool
from clarifai.client.user import User
from clarifai.utils.cli import AliasedGroup, display_co_resources, from_yaml, validate_context


@cli.group(['deployment', 'dp'], cls=AliasedGroup)
def deployment():
  """Manage Deployments: create, delete, list"""


@deployment.command(['c'])
@click.argument('nodepool_id')
@click.argument('deployment_id')
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the deployment config file.')
@click.pass_context
def create(ctx, nodepool_id, deployment_id, config):
  """Create a new Deployment with the given config file."""

  validate_context(ctx)
  if not nodepool_id:
    deployment_config = from_yaml(config)
    nodepool_id = deployment_config['deployment']['nodepools'][0]['id']

  nodepool = Nodepool(
      nodepool_id=nodepool_id,
      user_id=ctx.obj.current.user_id,
      pat=ctx.obj.current.pat,
      base_url=ctx.obj.current.api_base)
  if deployment_id:
    nodepool.create_deployment(config, deployment_id=deployment_id)
  else:
    nodepool.create_deployment(config)


@deployment.command(['ls'])
@click.argument('nodepool_id', default="")
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=16)
@click.pass_context
def list(ctx, nodepool_id, page_no, per_page):
  """List all deployments for the nodepool."""

  validate_context(ctx)
  if nodepool_id:
    nodepool = Nodepool(
        nodepool_id=nodepool_id,
        user_id=ctx.obj.current.user_id,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base)
    response = nodepool.list_deployments(page_no=page_no, per_page=per_page)
  else:
    user = User(
        user_id=ctx.obj.current.user_id,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base)
    ccs = user.list_compute_clusters(page_no, per_page)
    nps = []
    for cc in ccs:
      compute_cluster = ComputeCluster(
          compute_cluster_id=cc.id,
          user_id=ctx.obj.current.user_id,
          pat=ctx.obj.current.pat,
          base_url=ctx.obj.current.api_base)
      nps.extend([i for i in compute_cluster.list_nodepools(page_no, per_page)])
    response = []
    for np in nps:
      nodepool = Nodepool(
          nodepool_id=np.id,
          user_id=ctx.obj.current.user_id,
          pat=ctx.obj.current.pat,
          base_url=ctx.obj.current.api_base)
      response.extend([i for i in nodepool.list_deployments(page_no=page_no, per_page=per_page)])

  display_co_resources(
      response,
      custom_columns={
          'ID': lambda c: c.id,
          'USER_ID': lambda c: c.user_id,
          'COMPUTE_CLUSTER_ID': lambda c: c.nodepools[0].compute_cluster.id,
          'NODEPOOL_ID': lambda c: c.nodepools[0].id,
          'MODEL_USER_ID': lambda c: c.worker.model.user_id,
          'MODEL_APP_ID': lambda c: c.worker.model.app_id,
          'MODEL_ID': lambda c: c.worker.model.id,
          'MODEL_VERSION_ID': lambda c: c.worker.model.model_version.id,
          'DESCRIPTION': lambda c: c.description,
      })


@deployment.command(['rm'])
@click.argument('nodepool_id')
@click.argument('deployment_id')
@click.pass_context
def delete(ctx, nodepool_id, deployment_id):
  """Deletes a deployment for the nodepool."""

  validate_context(ctx)
  nodepool = Nodepool(
      nodepool_id=nodepool_id,
      user_id=ctx.obj.current.user_id,
      pat=ctx.obj.current.pat,
      base_url=ctx.obj.current.api_base)
  nodepool.delete_deployments([deployment_id])



import importlib
import inspect
import os
import re
import sys
import tarfile
import time
from string import Template

import yaml
from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2
from google.protobuf import json_format
from rich import print
from rich.markup import escape

from clarifai.client import BaseClient
from clarifai.runners.models.model_class import ModelClass
from clarifai.runners.utils.const import (
    AVAILABLE_PYTHON_IMAGES, AVAILABLE_TORCH_IMAGES, CONCEPTS_REQUIRED_MODEL_TYPE,
    DEFAULT_DOWNLOAD_CHECKPOINT_WHEN, DEFAULT_PYTHON_VERSION, DEFAULT_RUNTIME_DOWNLOAD_PATH,
    PYTHON_BASE_IMAGE, TORCH_BASE_IMAGE)
from clarifai.runners.utils.loader import HuggingFaceLoader
from clarifai.urls.helper import ClarifaiUrlHelper
from clarifai.utils.logging import logger
from clarifai.versions import CLIENT_VERSION

# parse the user's requirements.txt to determine the proper base image to build on top of, based on the torch and other large dependencies and it's versions
# List of dependencies to look for
dependencies = [
    'torch',
    'clarifai',
]


def _clear_line(n: int = 1) -> None:
  LINE_UP = '\033[1A'  # Move cursor up one line
  LINE_CLEAR = '\x1b[2K'  # Clear the entire line
  for _ in range(n):
    print(LINE_UP, end=LINE_CLEAR, flush=True)


class ModelBuilder:
  DEFAULT_CHECKPOINT_SIZE = 50 * 1024**3  # 50 GiB

  def __init__(self, folder: str, validate_api_ids: bool = True, download_validation_only=False):
    """
    :param folder: The folder containing the model.py, config.yaml, requirements.txt and
    checkpoints.
    :param validate_api_ids: Whether to validate the user_id and app_id in the config file. TODO(zeiler):
    deprecate in favor of download_validation_only.
    :param download_validation_only: Whether to skip the API config validation. Set to True if
    just downloading a checkpoint.
    """
    self._client = None
    if not validate_api_ids:  # for backwards compatibility
      download_validation_only = True
    self.download_validation_only = download_validation_only
    self.folder = self._validate_folder(folder)
    self.config = self._load_config(os.path.join(self.folder, 'config.yaml'))
    self._validate_config()
    self.model_proto = self._get_model_proto()
    self.model_id = self.model_proto.id
    self.model_version_id = None
    self.inference_compute_info = self._get_inference_compute_info()
    self.is_v3 = True  # Do model build for v3

  def create_model_instance(self, load_model=True):
    """
    Create an instance of the model class, as specified in the config file.
    """
    # look for default model.py file location
    for loc in ["model.py", "1/model.py"]:
      model_file = os.path.join(self.folder, loc)
      if os.path.exists(model_file):
        break
    if not os.path.exists(model_file):
      raise Exception("Model file not found.")

    module_name = os.path.basename(model_file).replace(".py", "")

    spec = importlib.util.spec_from_file_location(module_name, model_file)
    module = importlib.util.module_from_spec(spec)
    sys.modules[module_name] = module
    spec.loader.exec_module(module)

    # Find all classes in the model.py file that are subclasses of ModelClass
    classes = [
        cls for _, cls in inspect.getmembers(module, inspect.isclass)
        if issubclass(cls, ModelClass) and cls.__module__ == module.__name__
    ]
    #  Ensure there is exactly one subclass of BaseRunner in the model.py file
    if len(classes) != 1:
      # check for old inheritence structure, ModelRunner used to be a ModelClass
      runner_classes = [
          cls for _, cls in inspect.getmembers(module, inspect.isclass)
          if cls.__module__ == module.__name__ and any(c.__name__ == 'ModelRunner'
                                                       for c in cls.__bases__)
      ]
      if runner_classes and len(runner_classes) == 1:
        raise Exception(
            f'Could not determine model class.'
            f' Models should now inherit from {ModelClass.__module__}.ModelClass, not ModelRunner.'
            f' Please update your model "{runner_classes[0].__name__}" to inherit from ModelClass.'
        )
      raise Exception(
          "Could not determine model class. There should be exactly one model inheriting from ModelClass defined in the model.py"
      )
    model_class = classes[0]

    # initialize the model
    model = model_class()
    if load_model:
      model.load_model()
    return model

  def _validate_folder(self, folder):
    if folder == ".":
      folder = ""  # will getcwd() next which ends with /
    if not os.path.isabs(folder):
      folder = os.path.join(os.getcwd(), folder)
    logger.info(f"Validating folder: {folder}")
    if not os.path.exists(folder):
      raise FileNotFoundError(f"Folder {folder} not found, please provide a valid folder path")
    files = os.listdir(folder)
    assert "config.yaml" in files, "config.yaml not found in the folder"
    # If just downloading we don't need requirements.txt or the python code, we do need the
    # 1/ folder to put 1/checkpoints into.
    assert "1" in files, "Subfolder '1' not found in the folder"
    if not self.download_validation_only:
      assert "requirements.txt" in files, "requirements.txt not found in the folder"
      subfolder_files = os.listdir(os.path.join(folder, '1'))
      assert 'model.py' in subfolder_files, "model.py not found in the folder"
    return folder

  @staticmethod
  def _load_config(config_file: str):
    with open(config_file, 'r') as file:
      config = yaml.safe_load(file)
    return config

  def _validate_config_checkpoints(self):
    """
    Validates the checkpoints section in the config file.
      return loader_type, repo_id, hf_token, when, allowed_file_patterns, ignore_file_patterns
    :return: loader_type the type of loader or None if no checkpoints.
    :return: repo_id location of checkpoint.
    :return: hf_token token to access checkpoint.
    :return: when one of ['upload', 'build', 'runtime'] to download checkpoint
    :return: allowed_file_patterns patterns to allow in downloaded checkpoint
    :return: ignore_file_patterns patterns to ignore in downloaded checkpoint
    """
    if "checkpoints" not in self.config:
      return None, None, None, DEFAULT_DOWNLOAD_CHECKPOINT_WHEN, None, None
    assert "type" in self.config.get("checkpoints"), "No loader type specified in the config file"
    loader_type = self.config.get("checkpoints").get("type")
    if not loader_type:
      logger.info("No loader type specified in the config file for checkpoints")
      return None, None, None, DEFAULT_DOWNLOAD_CHECKPOINT_WHEN, None, None
    checkpoints = self.config.get("checkpoints")
    if 'when' not in checkpoints:
      logger.warn(
          f"No 'when' specified in the config file for checkpoints, defaulting to download at {DEFAULT_DOWNLOAD_CHECKPOINT_WHEN}"
      )
    when = checkpoints.get("when", DEFAULT_DOWNLOAD_CHECKPOINT_WHEN)
    assert when in [
        "upload",
        "build",
        "runtime",
    ], "Invalid value for when in the checkpoint loader when, needs to be one of ['upload', 'build', 'runtime']"
    assert loader_type == "huggingface", "Only huggingface loader supported for now"
    if loader_type == "huggingface":
      assert "repo_id" in self.config.get("checkpoints"), "No repo_id specified in the config file"
      repo_id = self.config.get("checkpoints").get("repo_id")

      # get from config.yaml otherwise fall back to HF_TOKEN env var.
      hf_token = self.config.get("checkpoints").get("hf_token", os.environ.get("HF_TOKEN", None))

      allowed_file_patterns = self.config.get("checkpoints").get('allowed_file_patterns', None)
      if isinstance(allowed_file_patterns, str):
        allowed_file_patterns = [allowed_file_patterns]
      ignore_file_patterns = self.config.get("checkpoints").get('ignore_file_patterns', None)
      if isinstance(ignore_file_patterns, str):
        ignore_file_patterns = [ignore_file_patterns]
      return loader_type, repo_id, hf_token, when, allowed_file_patterns, ignore_file_patterns

  def _check_app_exists(self):
    resp = self.client.STUB.GetApp(service_pb2.GetAppRequest(user_app_id=self.client.user_app_id))
    if resp.status.code == status_code_pb2.SUCCESS:
      return True
    if resp.status.code == status_code_pb2.CONN_KEY_INVALID:
      logger.error(
          f"Invalid PAT provided for user {self.client.user_app_id.user_id}. Please check your PAT and try again."
      )
      return False
    logger.error(
        f"Error checking API {self._base_api} for user app {self.client.user_app_id.user_id}/{self.client.user_app_id.app_id}. Error code: {resp.status.code}"
    )
    logger.error(
        f"App {self.client.user_app_id.app_id} not found for user {self.client.user_app_id.user_id}. Please create the app first and try again."
    )
    return False

  def _validate_config_model(self):
    assert "model" in self.config, "model section not found in the config file"
    model = self.config.get('model')
    assert "user_id" in model, "user_id not found in the config file"
    assert "app_id" in model, "app_id not found in the config file"
    assert "model_type_id" in model, "model_type_id not found in the config file"
    assert "id" in model, "model_id not found in the config file"
    if '.' in model.get('id'):
      logger.error(
          "Model ID cannot contain '.', please remove it from the model_id in the config file")
      sys.exit(1)

    assert model.get('user_id') != "", "user_id cannot be empty in the config file"
    assert model.get('app_id') != "", "app_id cannot be empty in the config file"
    assert model.get('model_type_id') != "", "model_type_id cannot be empty in the config file"
    assert model.get('id') != "", "model_id cannot be empty in the config file"

    if not self._check_app_exists():
      sys.exit(1)

  def _validate_config(self):
    if not self.download_validation_only:
      self._validate_config_model()

      assert "inference_compute_info" in self.config, "inference_compute_info not found in the config file"

      if self.config.get("concepts"):
        model_type_id = self.config.get('model').get('model_type_id')
        assert model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE, f"Model type {model_type_id} not supported for concepts"

    if self.config.get("checkpoints"):
      loader_type, _, hf_token, _, _, _ = self._validate_config_checkpoints()

      if loader_type == "huggingface" and hf_token:
        is_valid_token = HuggingFaceLoader.validate_hftoken(hf_token)
        if not is_valid_token:
          logger.error(
              "Invalid Hugging Face token provided in the config file, this might cause issues with downloading the restricted model checkpoints."
          )
          logger.info("Continuing without Hugging Face token")

    num_threads = self.config.get("num_threads")
    if num_threads or num_threads == 0:
      assert isinstance(num_threads, int) and num_threads >= 1, ValueError(
          f"`num_threads` must be an integer greater than or equal to 1. Received type {type(num_threads)} with value {num_threads}."
      )
    else:
      num_threads = int(os.environ.get("CLARIFAI_NUM_THREADS", 16))
      self.config["num_threads"] = num_threads

  @staticmethod
  def _get_tar_file_content_size(tar_file_path):
    """
    Calculates the total size of the contents of a tar file.

    Args:
      tar_file_path (str): The path to the tar file.

    Returns:
      int: The total size of the contents in bytes.
    """
    total_size = 0
    with tarfile.open(tar_file_path, 'r') as tar:
      for member in tar:
        if member.isfile():
          total_size += member.size
    return total_size

  @property
  def client(self):
    if self._client is None:
      assert "model" in self.config, "model info not found in the config file"
      model = self.config.get('model')
      assert "user_id" in model, "user_id not found in the config file"
      assert "app_id" in model, "app_id not found in the config file"
      # The owner of the model and the app.
      user_id = model.get('user_id')
      app_id = model.get('app_id')

      self._base_api = os.environ.get('CLARIFAI_API_BASE', 'https://api.clarifai.com')
      self._client = BaseClient(user_id=user_id, app_id=app_id, base=self._base_api)

    return self._client

  @property
  def model_url(self):
    url_helper = ClarifaiUrlHelper(self._client.auth_helper)
    if self.model_version_id is not None:
      return url_helper.clarifai_url(self.client.user_app_id.user_id,
                                     self.client.user_app_id.app_id, "models", self.model_id)
    else:
      return url_helper.clarifai_url(self.client.user_app_id.user_id,
                                     self.client.user_app_id.app_id, "models", self.model_id,
                                     self.model_version_id)

  def _get_model_proto(self):
    assert "model" in self.config, "model info not found in the config file"
    model = self.config.get('model')

    assert "model_type_id" in model, "model_type_id not found in the config file"
    assert "id" in model, "model_id not found in the config file"
    if not self.download_validation_only:
      assert "user_id" in model, "user_id not found in the config file"
      assert "app_id" in model, "app_id not found in the config file"

    model_proto = json_format.ParseDict(model, resources_pb2.Model())

    return model_proto

  def _get_inference_compute_info(self):
    assert ("inference_compute_info" in self.config
           ), "inference_compute_info not found in the config file"
    inference_compute_info = self.config.get('inference_compute_info')
    return json_format.ParseDict(inference_compute_info, resources_pb2.ComputeInfo())

  def check_model_exists(self):
    resp = self.client.STUB.GetModel(
        service_pb2.GetModelRequest(
            user_app_id=self.client.user_app_id, model_id=self.model_proto.id))
    if resp.status.code == status_code_pb2.SUCCESS:
      return True
    return False

  def maybe_create_model(self):
    if self.check_model_exists():
      logger.info(
          f"Model '{self.client.user_app_id.user_id}/{self.client.user_app_id.app_id}/models/{self.model_proto.id}' already exists, "
          f"will create a new version for it.")
      return

    request = service_pb2.PostModelsRequest(
        user_app_id=self.client.user_app_id,
        models=[self.model_proto],
    )
    return self.client.STUB.PostModels(request)

  def _match_req_line(self, line):
    line = line.strip()
    if not line or line.startswith('#'):
      return None, None
    # split on whitespace followed by #
    line = re.split(r'\s+#', line)[0]
    if "==" in line:
      pkg, version = line.split("==")
    elif ">=" in line:
      pkg, version = line.split(">=")
    elif ">" in line:
      pkg, version = line.split(">")
    elif "<=" in line:
      pkg, version = line.split("<=")
    elif "<" in line:
      pkg, version = line.split("<")
    else:
      pkg, version = line, None  # No version specified
    for dep in dependencies:
      if dep == pkg:
        if dep == 'torch' and line.find(
            'whl/cpu') > 0:  # Ignore torch-cpu whl files, use base mage.
          return None, None
        return dep.strip(), version.strip() if version else None
    return None, None

  def _parse_requirements(self):
    dependencies_version = {}
    with open(os.path.join(self.folder, 'requirements.txt'), 'r') as file:
      for line in file:
        # Skip empty lines and comments
        dependency, version = self._match_req_line(line)
        if dependency is None:
          continue
        dependencies_version[dependency] = version if version else None
    return dependencies_version

  def create_dockerfile(self):
    dockerfile_template = os.path.join(
        os.path.dirname(os.path.dirname(__file__)),
        'dockerfile_template',
        'Dockerfile.template',
    )

    with open(dockerfile_template, 'r') as template_file:
      dockerfile_template = template_file.read()

    dockerfile_template = Template(dockerfile_template)

    # Get the Python version from the config file
    build_info = self.config.get('build_info', {})
    if 'python_version' in build_info:
      python_version = build_info['python_version']
      if python_version not in AVAILABLE_PYTHON_IMAGES:
        raise Exception(
            f"Python version {python_version} not supported, please use one of the following versions: {AVAILABLE_PYTHON_IMAGES} in your config.yaml"
        )

      logger.info(
          f"Using Python version {python_version} from the config file to build the Dockerfile")
    else:
      logger.info(
          f"Python version not found in the config file, using default Python version: {DEFAULT_PYTHON_VERSION}"
      )
      python_version = DEFAULT_PYTHON_VERSION

    # This is always the final image used for runtime.
    final_image = PYTHON_BASE_IMAGE.format(python_version=python_version)
    downloader_image = PYTHON_BASE_IMAGE.format(python_version=python_version)

    # Parse the requirements.txt file to determine the base image
    dependencies = self._parse_requirements()
    if 'torch' in dependencies and dependencies['torch']:
      torch_version = dependencies['torch']

      # Sort in reverse so that newer cuda versions come first and are preferred.
      for image in sorted(AVAILABLE_TORCH_IMAGES, reverse=True):
        if torch_version in image and f'py{python_version}' in image:
          # like cu124, rocm6.3, etc.
          gpu_version = image.split('-')[-1]
          final_image = TORCH_BASE_IMAGE.format(
              torch_version=torch_version,
              python_version=python_version,
              gpu_version=gpu_version,
          )
          logger.info(f"Using Torch version {torch_version} base image to build the Docker image")
          break

    if 'clarifai' not in dependencies:
      raise Exception(
          f"clarifai not found in requirements.txt, please add clarifai to the requirements.txt file with a fixed version. Current version is clarifai=={CLIENT_VERSION}"
      )
    clarifai_version = dependencies['clarifai']
    if not clarifai_version:
      logger.warn(
          f"clarifai version not found in requirements.txt, using the latest version {CLIENT_VERSION}"
      )
      clarifai_version = CLIENT_VERSION
      lines = []
      with open(os.path.join(self.folder, 'requirements.txt'), 'r') as file:
        for line in file:
          # if the line without whitespace is "clarifai"
          dependency, version = self._match_req_line(line)
          if dependency and dependency == "clarifai":
            lines.append(line.replace("clarifai", f"clarifai=={CLIENT_VERSION}"))
          else:
            lines.append(line)
      with open(os.path.join(self.folder, 'requirements.txt'), 'w') as file:
        file.writelines(lines)
      logger.warn(f"Updated requirements.txt to have clarifai=={CLIENT_VERSION}")

    # Replace placeholders with actual values
    dockerfile_content = dockerfile_template.safe_substitute(
        name='main',
        FINAL_IMAGE=final_image,  # for pip requirements
        DOWNLOADER_IMAGE=downloader_image,  # for downloading checkpoints
        CLARIFAI_VERSION=clarifai_version,  # for clarifai
    )

    # Write Dockerfile
    with open(os.path.join(self.folder, 'Dockerfile'), 'w') as dockerfile:
      dockerfile.write(dockerfile_content)

  @property
  def checkpoint_path(self):
    return self._checkpoint_path(self.folder)

  def _checkpoint_path(self, folder):
    return os.path.join(folder, self.checkpoint_suffix)

  @property
  def checkpoint_suffix(self):
    return os.path.join('1', 'checkpoints')

  @property
  def tar_file(self):
    return f"{self.folder}.tar.gz"

  def default_runtime_checkpoint_path(self):
    return DEFAULT_RUNTIME_DOWNLOAD_PATH

  def download_checkpoints(self,
                           stage: str = DEFAULT_DOWNLOAD_CHECKPOINT_WHEN,
                           checkpoint_path_override: str = None):
    """
    Downloads the checkpoints specified in the config file.

    :param stage: The stage of the build process. This is used to determine when to download the
    checkpoints. The stage can be one of ['build', 'upload', 'runtime']. If you want to force
    downloading now then set stage to match e when field of the checkpoints section of you config.yaml.
    :param checkpoint_path_override: The path to download the checkpoints to (with 1/checkpoints added as suffix). If not provided, the
    default path is used based on the folder ModelUploader was initialized with. The checkpoint_suffix will be appended to the path.
    If stage is 'runtime' and checkpoint_path_override is None, the default runtime path will be used.

    :return: The path to the downloaded checkpoints. Even if it doesn't download anything, it will return the default path.
    """
    path = self.checkpoint_path  # default checkpoint path.
    if not self.config.get("checkpoints"):
      logger.info("No checkpoints specified in the config file")
      return path
    clarifai_model_type_id = self.config.get('model').get('model_type_id')

    loader_type, repo_id, hf_token, when, allowed_file_patterns, ignore_file_patterns = self._validate_config_checkpoints(
    )
    if stage not in ["build", "upload", "runtime"]:
      raise Exception("Invalid stage provided, must be one of ['build', 'upload', 'runtime']")
    if when != stage:
      logger.info(
          f"Skipping downloading checkpoints for stage {stage} since config.yaml says to download them at stage {when}"
      )
      return path

    success = False
    if loader_type == "huggingface":
      loader = HuggingFaceLoader(
          repo_id=repo_id, token=hf_token, model_type_id=clarifai_model_type_id)
      # for runtime default to /tmp path
      if stage == "runtime" and checkpoint_path_override is None:
        checkpoint_path_override = self.default_runtime_checkpoint_path()
      path = checkpoint_path_override if checkpoint_path_override else self.checkpoint_path
      success = loader.download_checkpoints(
          path,
          allowed_file_patterns=allowed_file_patterns,
          ignore_file_patterns=ignore_file_patterns)

    if loader_type:
      if not success:
        logger.error(f"Failed to download checkpoints for model {repo_id}")
        sys.exit(1)
      else:
        logger.info(f"Downloaded checkpoints for model {repo_id}")
    return path

  def _concepts_protos_from_concepts(self, concepts):
    concept_protos = []
    for concept in concepts:
      concept_protos.append(resources_pb2.Concept(
          id=str(concept[0]),
          name=concept[1],
      ))
    return concept_protos

  def hf_labels_to_config(self, labels, config_file):
    with open(config_file, 'r') as file:
      config = yaml.safe_load(file)
    model = config.get('model')
    model_type_id = model.get('model_type_id')
    assert model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE, f"Model type {model_type_id} not supported for concepts"
    concept_protos = self._concepts_protos_from_concepts(labels)

    config['concepts'] = [{'id': concept.id, 'name': concept.name} for concept in concept_protos]

    with open(config_file, 'w') as file:
      yaml.dump(config, file, sort_keys=False)
    concepts = config.get('concepts')
    logger.info(f"Updated config.yaml with {len(concepts)} concepts.")

  def get_model_version_proto(self):

    model_version_proto = resources_pb2.ModelVersion(
        pretrained_model_config=resources_pb2.PretrainedModelConfig(),
        inference_compute_info=self.inference_compute_info,
    )

    model_type_id = self.config.get('model').get('model_type_id')
    if model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE:

      if 'concepts' in self.config:
        labels = self.config.get('concepts')
        logger.info(f"Found {len(labels)} concepts in the config file.")
        for concept in labels:
          concept_proto = json_format.ParseDict(concept, resources_pb2.Concept())
          model_version_proto.output_info.data.concepts.append(concept_proto)
      elif self.config.get("checkpoints") and HuggingFaceLoader.validate_concept(
          self.checkpoint_path):
        labels = HuggingFaceLoader.fetch_labels(self.checkpoint_path)
        logger.info(f"Found {len(labels)} concepts from the model checkpoints.")
        # sort the concepts by id and then update the config file
        labels = sorted(labels.items(), key=lambda x: int(x[0]))

        config_file = os.path.join(self.folder, 'config.yaml')
        try:
          self.hf_labels_to_config(labels, config_file)
        except Exception as e:
          logger.error(f"Failed to update the config.yaml file with the concepts: {e}")

        model_version_proto.output_info.data.concepts.extend(
            self._concepts_protos_from_concepts(labels))
    return model_version_proto

  def upload_model_version(self):
    file_path = f"{self.folder}.tar.gz"
    logger.debug(f"Will tar it into file: {file_path}")

    model_type_id = self.config.get('model').get('model_type_id')
    loader_type, repo_id, hf_token, when, _, _ = self._validate_config_checkpoints()

    if (model_type_id in CONCEPTS_REQUIRED_MODEL_TYPE) and 'concepts' not in self.config:
      logger.info(
          f"Model type {model_type_id} requires concepts to be specified in the config.yaml file.."
      )
      if self.config.get("checkpoints"):
        logger.info(
            "Checkpoints specified in the config.yaml file, will download the HF model's config.json file to infer the concepts."
        )
        # If we don't already have the concepts, download the config.json file from HuggingFace
        if loader_type == "huggingface":
          # If the config.yaml says we'll download in the future (build time or runtime) then we need to get this config now.
          if when != "upload" and not HuggingFaceLoader.validate_config(self.checkpoint_path):
            input(
                "Press Enter to download the HuggingFace model's config.json file to infer the concepts and continue..."
            )
            loader = HuggingFaceLoader(repo_id=repo_id, token=hf_token)
            loader.download_config(self.checkpoint_path)

      else:
        logger.error(
            "No checkpoints specified in the config.yaml file to infer the concepts. Please either specify the concepts directly in the config.yaml file or include a checkpoints section to download the HF model's config.json file to infer the concepts."
        )
        return

    model_version_proto = self.get_model_version_proto()

    def filter_func(tarinfo):
      name = tarinfo.name
      exclude = [self.tar_file, "*~", "*.pyc", "*.pyo", "__pycache__"]
      if when != "upload":
        exclude.append(self.checkpoint_suffix)
      return None if any(name.endswith(ex) for ex in exclude) else tarinfo

    with tarfile.open(self.tar_file, "w:gz") as tar:
      tar.add(self.folder, arcname=".", filter=filter_func)
    logger.debug("Tarring complete, about to start upload.")

    file_size = os.path.getsize(self.tar_file)
    logger.debug(f"Size of the tar is: {file_size} bytes")

    self.storage_request_size = self._get_tar_file_content_size(file_path)
    if when != "upload" and self.config.get("checkpoints"):
      # Get the checkpoint size to add to the storage request.
      # First check for the env variable, then try querying huggingface. If all else fails, use the default.
      checkpoint_size = os.environ.get('CHECKPOINT_SIZE_BYTES', 0)
      if not checkpoint_size:
        _, repo_id, _, _, _, _ = self._validate_config_checkpoints()
        checkpoint_size = HuggingFaceLoader.get_huggingface_checkpoint_total_size(repo_id)
      if not checkpoint_size:
        checkpoint_size = self.DEFAULT_CHECKPOINT_SIZE
      self.storage_request_size += checkpoint_size

    resp = self.maybe_create_model()
    if not self.check_model_exists():
      logger.error(f"Failed to create model: {self.model_proto.id}. Details: {resp}")
      sys.exit(1)

    for response in self.client.STUB.PostModelVersionsUpload(
        self.model_version_stream_upload_iterator(model_version_proto, file_path),):
      percent_completed = 0
      if response.status.code == status_code_pb2.UPLOAD_IN_PROGRESS:
        percent_completed = response.status.percent_completed
      details = response.status.details

      _clear_line()
      print(
          f"Status: {response.status.description}, "
          f"Progress: {percent_completed}% - {details} ",
          f"request_id: {response.status.req_id}",
          end='\r',
          flush=True)
    if response.status.code != status_code_pb2.MODEL_BUILDING:
      logger.error(f"Failed to upload model version: {response}")
      return
    self.model_version_id = response.model_version_id
    logger.info(f"Created Model Version ID: {self.model_version_id}")
    logger.info(f"Full url to that version is: {self.model_url}")
    try:
      self.monitor_model_build()
    finally:
      if os.path.exists(self.tar_file):
        logger.debug(f"Cleaning up upload file: {self.tar_file}")
        os.remove(self.tar_file)

  def model_version_stream_upload_iterator(self, model_version_proto, file_path):
    yield self.init_upload_model_version(model_version_proto, file_path)
    with open(file_path, "rb") as f:
      file_size = os.path.getsize(file_path)
      chunk_size = int(127 * 1024 * 1024)  # 127MB chunk size
      num_chunks = (file_size // chunk_size) + 1
      logger.info("Uploading file...")
      logger.debug(f"File size: {file_size}")
      logger.debug(f"Chunk size: {chunk_size}")
      logger.debug(f"Number of chunks: {num_chunks}")
      read_so_far = 0
      for part_id in range(num_chunks):
        try:
          chunk_size = min(chunk_size, file_size - read_so_far)
          chunk = f.read(chunk_size)
          if not chunk:
            break
          read_so_far += len(chunk)
          yield service_pb2.PostModelVersionsUploadRequest(
              content_part=resources_pb2.UploadContentPart(
                  data=chunk,
                  part_number=part_id + 1,
                  range_start=read_so_far,
              ))
        except Exception as e:
          logger.exception(f"\nError uploading file: {e}")
          break

    if read_so_far == file_size:
      logger.info("Upload complete!")

  def init_upload_model_version(self, model_version_proto, file_path):
    file_size = os.path.getsize(file_path)
    logger.debug(f"Uploading model version of model {self.model_proto.id}")
    logger.debug(f"Using file '{os.path.basename(file_path)}' of size: {file_size} bytes")
    result = service_pb2.PostModelVersionsUploadRequest(
        upload_config=service_pb2.PostModelVersionsUploadConfig(
            user_app_id=self.client.user_app_id,
            model_id=self.model_proto.id,
            model_version=model_version_proto,
            total_size=file_size,
            storage_request_size=self.storage_request_size,
            is_v3=self.is_v3,
        ))
    return result

  def get_model_build_logs(self):
    logs_request = service_pb2.ListLogEntriesRequest(
        log_type="builder",
        user_app_id=self.client.user_app_id,
        model_id=self.model_proto.id,
        model_version_id=self.model_version_id,
        page=1,
        per_page=50)
    response = self.client.STUB.ListLogEntries(logs_request)

    return response

  def monitor_model_build(self):
    st = time.time()
    seen_logs = set()  # To avoid duplicate log messages
    while True:
      resp = self.client.STUB.GetModelVersion(
          service_pb2.GetModelVersionRequest(
              user_app_id=self.client.user_app_id,
              model_id=self.model_proto.id,
              version_id=self.model_version_id,
          ))
      status_code = resp.model_version.status.code
      logs = self.get_model_build_logs()
      for log_entry in logs.log_entries:
        if log_entry.url not in seen_logs:
          seen_logs.add(log_entry.url)
          logger.info(f"{escape(log_entry.message.strip())}")
      if status_code == status_code_pb2.MODEL_BUILDING:
        print(f"Model is building... (elapsed {time.time() - st:.1f}s)", end='\r', flush=True)

        # Fetch and display the logs
        time.sleep(1)
      elif status_code == status_code_pb2.MODEL_TRAINED:
        logger.info("Model build complete!")
        logger.info(f"Build time elapsed {time.time() - st:.1f}s)")
        logger.info(f"Check out the model at {self.model_url} version: {self.model_version_id}")
        return True
      else:
        logger.info(
            f"\nModel build failed with status: {resp.model_version.status} and response {resp}")
        return False


def upload_model(folder, stage, skip_dockerfile):
  """
  Uploads a model to Clarifai.

  :param folder: The folder containing the model files.
  :param stage: The stage we are calling download checkpoints from. Typically this would "upload" and will download checkpoints if config.yaml checkpoints section has when set to "upload". Other options include "runtime" to be used in load_model or "upload" to be used during model upload. Set this stage to whatever you have in config.yaml to force downloading now.
  :param skip_dockerfile: If True, will not create a Dockerfile.
  """
  builder = ModelBuilder(folder)
  builder.download_checkpoints(stage=stage)
  if not skip_dockerfile:
    builder.create_dockerfile()
  exists = builder.check_model_exists()
  if exists:
    logger.info(
        f"Model already exists at {builder.model_url}, this upload will create a new version for it."
    )
  else:
    logger.info(f"New model will be created at {builder.model_url} with it's first version.")

  input("Press Enter to continue...")
  builder.upload_model_version()



import click

from clarifai.cli.base import cli
from clarifai.client.compute_cluster import ComputeCluster
from clarifai.client.user import User
from clarifai.utils.cli import (AliasedGroup, display_co_resources, dump_yaml, from_yaml,
                                validate_context)


@cli.group(['nodepool', 'np'], cls=AliasedGroup)
def nodepool():
  """Manage Nodepools: create, delete, list"""


@nodepool.command(['c'])
@click.argument('compute_cluster_id')
@click.argument('nodepool_id')
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=True,
    help='Path to the nodepool config file.')
@click.pass_context
def create(ctx, compute_cluster_id, nodepool_id, config):
  """Create a new Nodepool with the given config file."""

  validate_context(ctx)
  nodepool_config = from_yaml(config)
  if not compute_cluster_id:
    if 'compute_cluster' not in nodepool_config['nodepool']:
      click.echo(
          "Please provide a compute cluster ID either in the config file or using --compute_cluster_id flag",
          err=True)
      return
    compute_cluster_id = nodepool_config['nodepool']['compute_cluster']['id']
  else:
    if 'compute_cluster' not in nodepool_config['nodepool']:
      nodepool_config['nodepool']['compute_cluster']['id'] = compute_cluster_id
      dump_yaml(config, nodepool_config)

  compute_cluster = ComputeCluster(
      compute_cluster_id=compute_cluster_id,
      user_id=ctx.obj.current.user_id,
      pat=ctx.obj.current.pat,
      base_url=ctx.obj.current.api_base)
  if nodepool_id:
    compute_cluster.create_nodepool(config, nodepool_id=nodepool_id)
  else:
    compute_cluster.create_nodepool(config)


@nodepool.command(['ls'])
@click.argument('compute_cluster_id', default="")
@click.option('--page_no', required=False, help='Page number to list.', default=1)
@click.option('--per_page', required=False, help='Number of items per page.', default=128)
@click.pass_context
def list(ctx, compute_cluster_id, page_no, per_page):
  """List all nodepools for the user across all compute clusters. If compute_cluster_id is provided
  it will list only within that compute cluster. """

  validate_context(ctx)

  cc_id = compute_cluster_id

  if cc_id:
    compute_cluster = ComputeCluster(
        compute_cluster_id=cc_id,
        user_id=ctx.obj.current.user_id,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base)
    response = compute_cluster.list_nodepools(page_no, per_page)
  else:
    user = User(
        user_id=ctx.obj.current.user_id,
        pat=ctx.obj.current.pat,
        base_url=ctx.obj.current.api_base)
    ccs = user.list_compute_clusters(page_no, per_page)
    response = []
    for cc in ccs:
      compute_cluster = ComputeCluster(
          compute_cluster_id=cc.id,
          user_id=ctx.obj.current.user_id,
          pat=ctx.obj.current.pat,
          base_url=ctx.obj.current.api_base)
      response.extend([i for i in compute_cluster.list_nodepools(page_no, per_page)])

  display_co_resources(
      response,
      custom_columns={
          'ID': lambda c: c.id,
          'USER_ID': lambda c: c.compute_cluster.user_id,
          'COMPUTE_CLUSTER_ID': lambda c: c.compute_cluster.id,
          'DESCRIPTION': lambda c: c.description,
      })


@nodepool.command(['rm'])
@click.argument('compute_cluster_id')
@click.argument('nodepool_id')
@click.pass_context
def delete(ctx, compute_cluster_id, nodepool_id):
  """Deletes a nodepool for the user."""

  validate_context(ctx)
  compute_cluster = ComputeCluster(
      compute_cluster_id=compute_cluster_id,
      user_id=ctx.obj.current.user_id,
      pat=ctx.obj.current.pat,
      base_url=ctx.obj.current.api_base)
  compute_cluster.delete_nodepools([nodepool_id])



import click

from clarifai.cli.base import cli


@cli.group(['model'])
def model():
  """Manage models: upload, test, local dev, predict, etc"""


@model.command()
@click.argument("model_path", type=click.Path(exists=True), required=False, default=".")
@click.option(
    '--stage',
    required=False,
    type=click.Choice(['runtime', 'build', 'upload'], case_sensitive=True),
    default="upload",
    show_default=True,
    help=
    'The stage we are calling download checkpoints from. Typically this would "upload" and will download checkpoints if config.yaml checkpoints section has when set to "upload". Other options include "runtime" to be used in load_model or "upload" to be used during model upload. Set this stage to whatever you have in config.yaml to force downloading now.'
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help=
    'Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile.',
)
def upload(model_path, stage, skip_dockerfile):
  """Upload a model to Clarifai.

    MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
  """
  from clarifai.runners.models.model_builder import upload_model
  upload_model(model_path, stage, skip_dockerfile)


@model.command()
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--out_path',
    type=click.Path(exists=False),
    required=False,
    default=None,
    help=
    'Option path to write the checkpoints to. This will place them in {out_path}/1/checkpoints If not provided it will default to {model_path}/1/checkpoints where the config.yaml is read.'
)
@click.option(
    '--stage',
    required=False,
    type=click.Choice(['runtime', 'build', 'upload'], case_sensitive=True),
    default="build",
    show_default=True,
    help=
    'The stage we are calling download checkpoints from. Typically this would be in the build stage which is the default. Other options include "runtime" to be used in load_model or "upload" to be used during model upload. Set this stage to whatever you have in config.yaml to force downloading now.'
)
def download_checkpoints(model_path, out_path, stage):
  """Download checkpoints from external source to local model_path

  MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
  """

  from clarifai.runners.models.model_builder import ModelBuilder
  builder = ModelBuilder(model_path, download_validation_only=True)
  builder.download_checkpoints(stage=stage, checkpoint_path_override=out_path)


@model.command()
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--mode',
    type=click.Choice(['env', 'container'], case_sensitive=False),
    default='env',
    show_default=True,
    help=
    'Specify how to test the model locally: "env" for virtual environment or "container" for Docker container. Defaults to "env".'
)
@click.option(
    '--keep_env',
    is_flag=True,
    help=
    'Keep the virtual environment after testing the model locally (applicable for virtualenv mode). Defaults to False.'
)
@click.option(
    '--keep_image',
    is_flag=True,
    help=
    'Keep the Docker image after testing the model locally (applicable for container mode). Defaults to False.'
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help=
    'Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile. Apply for `--mode conatainer`.',
)
def test_locally(model_path, keep_env=False, keep_image=False, mode='env', skip_dockerfile=False):
  """Test model locally."""
  try:
    from clarifai.runners.models import model_run_locally
    if mode == 'env' and keep_image:
      raise ValueError("'keep_image' is applicable only for 'container' mode")
    if mode == 'container' and keep_env:
      raise ValueError("'keep_env' is applicable only for 'env' mode")

    if mode == "env":
      click.echo("Testing model locally in a virtual environment...")
      model_run_locally.main(model_path, run_model_server=False, keep_env=keep_env)
    elif mode == "container":
      click.echo("Testing model locally inside a container...")
      model_run_locally.main(
          model_path,
          inside_container=True,
          run_model_server=False,
          keep_image=keep_image,
          skip_dockerfile=skip_dockerfile)
    click.echo("Model tested successfully.")
  except Exception as e:
    click.echo(f"Failed to test model locally: {e}", err=True)


@model.command()
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
@click.option(
    '--port',
    '-p',
    type=int,
    default=8000,
    show_default=True,
    help="The port to host the gRPC server for running the model locally. Defaults to 8000.")
@click.option(
    '--mode',
    type=click.Choice(['env', 'container'], case_sensitive=False),
    default='env',
    show_default=True,
    help=
    'Specifies how to run the model: "env" for virtual environment or "container" for Docker container. Defaults to "env".'
)
@click.option(
    '--keep_env',
    is_flag=True,
    help=
    'Keep the virtual environment after testing the model locally (applicable for virtualenv mode). Defaults to False.'
)
@click.option(
    '--keep_image',
    is_flag=True,
    help=
    'Keep the Docker image after testing the model locally (applicable for container mode). Defaults to False.'
)
@click.option(
    '--skip_dockerfile',
    is_flag=True,
    help=
    'Flag to skip generating a dockerfile so that you can manually edit an already created dockerfile. Apply for `--mode conatainer`.',
)
def run_locally(model_path, port, mode, keep_env, keep_image, skip_dockerfile=False):
  """Run the model locally and start a gRPC server to serve the model."""
  try:
    from clarifai.runners.models import model_run_locally
    if mode == 'env' and keep_image:
      raise ValueError("'keep_image' is applicable only for 'container' mode")
    if mode == 'container' and keep_env:
      raise ValueError("'keep_env' is applicable only for 'env' mode")

    if mode == "env":
      click.echo("Running model locally in a virtual environment...")
      model_run_locally.main(model_path, run_model_server=True, keep_env=keep_env, port=port)
    elif mode == "container":
      click.echo("Running model locally inside a container...")
      model_run_locally.main(
          model_path,
          inside_container=True,
          run_model_server=True,
          port=port,
          keep_image=keep_image,
          skip_dockerfile=skip_dockerfile)
    click.echo(f"Model server started locally from {model_path} in {mode} mode.")
  except Exception as e:
    click.echo(f"Failed to starts model server locally: {e}", err=True)


@model.command()
@click.argument(
    "model_path",
    type=click.Path(exists=True),
    required=False,
    default=".",
)
def local_dev(model_path):
  """Run the model as a local dev runner to help debug your model connected to the API. You must set several envvars such as CLARIFAI_PAT, CLARIFAI_RUNNER_ID, CLARIFAI_NODEPOOL_ID, CLARIFAI_COMPUTE_CLUSTER_ID.

  MODEL_PATH: Path to the model directory. If not specified, the current directory is used by default.
  """
  from clarifai.runners.server import serve
  serve(model_path)


@model.command()
@click.option(
    '--config',
    type=click.Path(exists=True),
    required=False,
    help='Path to the model predict config file.')
@click.option('--model_id', required=False, help='Model ID of the model used to predict.')
@click.option('--user_id', required=False, help='User ID of the model used to predict.')
@click.option('--app_id', required=False, help='App ID of the model used to predict.')
@click.option('--model_url', required=False, help='Model URL of the model used to predict.')
@click.option('--file_path', required=False, help='File path of file for the model to predict')
@click.option('--url', required=False, help='URL to the file for the model to predict')
@click.option('--bytes', required=False, help='Bytes to the file for the model to predict')
@click.option('--input_type', required=False, help='Type of input')
@click.option(
    '-cc_id',
    '--compute_cluster_id',
    required=False,
    help='Compute Cluster ID to use for the model')
@click.option('-np_id', '--nodepool_id', required=False, help='Nodepool ID to use for the model')
@click.option(
    '-dpl_id', '--deployment_id', required=False, help='Deployment ID to use for the model')
@click.option(
    '--inference_params', required=False, default='{}', help='Inference parameters to override')
@click.option('--output_config', required=False, default='{}', help='Output config to override')
@click.pass_context
def predict(ctx, config, model_id, user_id, app_id, model_url, file_path, url, bytes, input_type,
            compute_cluster_id, nodepool_id, deployment_id, inference_params, output_config):
  """Predict using the given model"""
  import json

  from clarifai.client.model import Model
  from clarifai.utils.cli import from_yaml, validate_context
  validate_context(ctx)
  if config:
    config = from_yaml(config)
    model_id, user_id, app_id, model_url, file_path, url, bytes, input_type, compute_cluster_id, nodepool_id, deployment_id, inference_params, output_config = (
        config.get(k, v)
        for k, v in [('model_id', model_id), ('user_id', user_id), ('app_id', app_id), (
            'model_url', model_url), ('file_path', file_path), ('url', url), ('bytes', bytes), (
                'input_type', input_type), ('compute_cluster_id', compute_cluster_id), (
                    'nodepool_id',
                    nodepool_id), ('deployment_id',
                                   deployment_id), ('inference_params',
                                                    inference_params), ('output_config',
                                                                        output_config)])
  if sum([opt[1] for opt in [(model_id, 1), (user_id, 1), (app_id, 1), (model_url, 3)]
          if opt[0]]) != 3:
    raise ValueError("Either --model_id & --user_id & --app_id or --model_url must be provided.")
  if compute_cluster_id or nodepool_id or deployment_id:
    if sum([
        opt[1] for opt in [(compute_cluster_id, 0.5), (nodepool_id, 0.5), (deployment_id, 1)]
        if opt[0]
    ]) != 1:
      raise ValueError(
          "Either --compute_cluster_id & --nodepool_id or --deployment_id must be provided.")
  if model_url:
    model = Model(url=model_url, pat=ctx.obj['pat'], base_url=ctx.obj['base_url'])
  else:
    model = Model(
        model_id=model_id,
        user_id=user_id,
        app_id=app_id,
        pat=ctx.obj['pat'],
        base_url=ctx.obj['base_url'])

  if inference_params:
    inference_params = json.loads(inference_params)
  if output_config:
    output_config = json.loads(output_config)

  if file_path:
    model_prediction = model.predict_by_filepath(
        filepath=file_path,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        inference_params=inference_params,
        output_config=output_config)
  elif url:
    model_prediction = model.predict_by_url(
        url=url,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        inference_params=inference_params,
        output_config=output_config)
  elif bytes:
    bytes = str.encode(bytes)
    model_prediction = model.predict_by_bytes(
        input_bytes=bytes,
        input_type=input_type,
        compute_cluster_id=compute_cluster_id,
        nodepool_id=nodepool_id,
        deployment_id=deployment_id,
        inference_params=inference_params,
        output_config=output_config)  ## TO DO: Add support for input_id
  click.echo(model_prediction)



from concurrent.futures import ThreadPoolExecutor
from typing import List, Tuple, Type

from clarifai_grpc.grpc.api import resources_pb2
from google.protobuf.struct_pb2 import Struct

from clarifai.client.input import Inputs
from clarifai.datasets.upload.base import ClarifaiDataLoader, ClarifaiDataset


class MultiModalDataset(ClarifaiDataset):

  def __init__(self,
               data_generator: Type[ClarifaiDataLoader],
               dataset_id: str,
               max_workers: int = 4) -> None:
    super().__init__(data_generator, dataset_id, max_workers)

  def _extract_protos(
      self,
      batch_input_ids: List[str],
  ) -> Tuple[List[resources_pb2.Input]]:
    """ Creats Multimodal (image and text) input protos for batch of input ids.
        Args:
            batch_input_ids: List of input IDs to retrieve the protos for.
        Returns:
            input_protos: List of input protos.

        """
    input_protos, annotation_protos = [], []

    def process_data_item(id):
      data_item = self.data_generator[id]
      metadata = Struct()
      image_bytes = data_item.image_bytes
      text = data_item.text
      labels = data_item.labels if ((data_item.labels is None) or
                                    isinstance(data_item.labels, list)) else [data_item.labels]
      input_id = f"{self.dataset_id}-{id}" if data_item.id is None else f"{self.dataset_id}-{str(data_item.id)}"
      if data_item.metadata is not None:
        metadata.update(data_item.metadata)
      else:
        metadata = None

      self.all_input_ids[id] = input_id
      if data_item.image_bytes is not None:
        input_protos.append(
            Inputs.get_input_from_bytes(
                input_id=input_id,
                image_bytes=image_bytes,
                dataset_id=self.dataset_id,
                labels=labels,
                metadata=metadata))
      else:
        input_protos.append(
            Inputs.get_text_input(
                input_id=input_id,
                raw_text=text,
                dataset_id=self.dataset_id,
                labels=labels,
                metadata=metadata))

    with ThreadPoolExecutor(max_workers=self.max_workers) as executor:
      futures = [executor.submit(process_data_item, id) for id in batch_input_ids]

      for job in futures:
        job.result()

    return input_protos, annotation_protos



import hashlib
import os
import platform
import shutil
import signal
import subprocess
import sys
import tempfile
import time
import traceback
import venv

from clarifai_grpc.grpc.api import resources_pb2, service_pb2
from clarifai_grpc.grpc.api.status import status_code_pb2, status_pb2

from clarifai.runners.models.model_builder import ModelBuilder
from clarifai.runners.utils.url_fetcher import ensure_urls_downloaded
from clarifai.utils.logging import logger


class ModelRunLocally:

  def __init__(self, model_path):
    self.model_path = model_path
    self.requirements_file = os.path.join(self.model_path, "requirements.txt")

    # ModelBuilder contains multiple useful methods to interact with the model
    self.builder = ModelBuilder(self.model_path, download_validation_only=True)
    self.config = self.builder.config

  def _requirements_hash(self):
    """Generate a hash of the requirements file."""
    with open(self.requirements_file, "r") as f:
      return hashlib.md5(f.read().encode('utf-8')).hexdigest()

  def _get_env_executable(self):
    """Get the python executable from the virtual environment."""
    # Depending on the platform, venv scripts are placed in either "Scripts" (Windows) or "bin" (Linux/Mac)
    if platform.system().lower().startswith("win"):
      scripts_folder = "Scripts"
      python_exe = "python.exe"
      pip_exe = "pip.exe"
    else:
      scripts_folder = "bin"
      python_exe = "python"
      pip_exe = "pip"

    self.python_executable = os.path.join(self.venv_dir, scripts_folder, python_exe)
    self.pip_executable = os.path.join(self.venv_dir, scripts_folder, pip_exe)

    return self.python_executable, self.pip_executable

  def create_temp_venv(self):
    """Create a temporary virtual environment."""
    requirements_hash = self._requirements_hash()

    temp_dir = os.path.join(tempfile.gettempdir(), str(requirements_hash))
    venv_dir = os.path.join(temp_dir, "venv")

    if os.path.exists(temp_dir):
      logger.info(f"Using previous virtual environment at {temp_dir}")
      use_existing_venv = True
    else:
      logger.info("Creating temporary virtual environment...")
      use_existing_venv = False
      venv.create(venv_dir, with_pip=True)
      logger.info(f"Created temporary virtual environment at {venv_dir}")

    self.venv_dir = venv_dir
    self.temp_dir = temp_dir
    self.python_executable, self.pip_executable = self._get_env_executable()

    return use_existing_venv

  def install_requirements(self):
    """Install the dependencies from requirements.txt and Clarifai."""
    _, pip_executable = self._get_env_executable()
    try:
      logger.info(
          f"Installing requirements from {self.requirements_file}... in the virtual environment {self.venv_dir}"
      )
      subprocess.check_call([pip_executable, "install", "-r", self.requirements_file])
      logger.info("Installing Clarifai package...")
      subprocess.check_call([pip_executable, "install", "clarifai"])
      logger.info("Requirements installed successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error installing requirements: {e}")
      self.clean_up()
      sys.exit(1)

  def _build_request(self):
    """Create a mock inference request for testing the model."""

    model_version_proto = self.builder.get_model_version_proto()
    model_version_proto.id = "model_version"

    return service_pb2.PostModelOutputsRequest(
        model=resources_pb2.Model(model_version=model_version_proto),
        inputs=[
            resources_pb2.Input(data=resources_pb2.Data(
                text=resources_pb2.Text(raw="How many people live in new york?"),
                image=resources_pb2.Image(url="https://samples.clarifai.com/metro-north.jpg"),
                audio=resources_pb2.Audio(url="https://samples.clarifai.com/GoodMorning.wav"),
                video=resources_pb2.Video(url="https://samples.clarifai.com/beer.mp4"),
            ))
        ],
    )

  def _build_stream_request(self):
    request = self._build_request()
    ensure_urls_downloaded(request)
    for i in range(1):
      yield request

  def _run_model_inference(self, model):
    """Perform inference using the model."""
    request = self._build_request()
    stream_request = self._build_stream_request()

    ensure_urls_downloaded(request)
    predict_response = None
    generate_response = None
    stream_response = None
    try:
      predict_response = model.predict(request)
    except NotImplementedError:
      logger.info("Model does not implement predict() method.")
    except Exception as e:
      logger.error(f"Model Prediction failed: {e}")
      traceback.print_exc()
      predict_response = service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_PREDICTION_FAILED,
          description="Prediction failed",
          details="",
          internal_details=str(e),
      ))

    if predict_response:
      if predict_response.outputs[0].status.code != status_code_pb2.SUCCESS:
        logger.error(f"Moddel Prediction failed: {predict_response}")
      else:
        logger.info(f"Model Prediction succeeded: {predict_response}")

    try:
      generate_response = model.generate(request)
    except NotImplementedError:
      logger.info("Model does not implement generate() method.")
    except Exception as e:
      logger.error(f"Model Generation failed: {e}")
      traceback.print_exc()
      generate_response = service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_GENERATION_FAILED,
          description="Generation failed",
          details="",
          internal_details=str(e),
      ))

    if generate_response:
      generate_first_res = next(generate_response)
      if generate_first_res.outputs[0].status.code != status_code_pb2.SUCCESS:
        logger.error(f"Moddel Prediction failed: {generate_first_res}")
      else:
        logger.info(
            f"Model Prediction succeeded for generate and first response: {generate_first_res}")

    try:
      stream_response = model.stream(stream_request)
    except NotImplementedError:
      logger.info("Model does not implement stream() method.")
    except Exception as e:
      logger.error(f"Model Stream failed: {e}")
      traceback.print_exc()
      stream_response = service_pb2.MultiOutputResponse(status=status_pb2.Status(
          code=status_code_pb2.MODEL_STREAM_FAILED,
          description="Stream failed",
          details="",
          internal_details=str(e),
      ))

    if stream_response:
      stream_first_res = next(stream_response)
      if stream_first_res.outputs[0].status.code != status_code_pb2.SUCCESS:
        logger.error(f"Moddel Prediction failed: {stream_first_res}")
      else:
        logger.info(
            f"Model Prediction succeeded for stream and first response: {stream_first_res}")

  def _run_test(self):
    """Test the model locally by making a prediction."""
    # Create the model
    model = self.builder.create_model_instance()
    # send an inference.
    self._run_model_inference(model)

  def test_model(self):
    """Test the model by running it locally in the virtual environment."""

    import_path = repr(os.path.dirname(os.path.abspath(__file__)))
    model_path = repr(self.model_path)

    command_string = (f"import sys; "
                      f"sys.path.append({import_path}); "
                      f"from model_run_locally import ModelRunLocally; "
                      f"ModelRunLocally({model_path})._run_test()")

    command = [self.python_executable, "-c", command_string]
    process = None
    try:
      logger.info("Testing the model locally...")
      process = subprocess.Popen(command)
      # Wait for the process to complete
      process.wait()
      if process.returncode == 0:
        logger.info("Model tested successfully!")
      if process.returncode != 0:
        raise subprocess.CalledProcessError(process.returncode, command)
    except subprocess.CalledProcessError as e:
      logger.error(f"Error testing the model: {e}")
      sys.exit(1)
    except Exception as e:
      logger.error(f"Unexpected error: {e}")
      sys.exit(1)
    finally:
      # After the function runs, check if the process is still running
      if process and process.poll() is None:
        logger.info("Process is still running. Terminating process.")
        process.terminate()
        try:
          process.wait(timeout=5)
        except subprocess.TimeoutExpired:
          logger.info("Process did not terminate gracefully. Killing process.")
          # Kill the process if it doesn't terminate after 5 seconds
          process.kill()

  # run the model server
  def run_model_server(self, port=8080):
    """Run the Clarifai Runners's model server."""

    command = [
        self.python_executable, "-m", "clarifai.runners.server", "--model_path", self.model_path,
        "--grpc", "--port",
        str(port)
    ]
    try:
      logger.info(
          f"Starting model server at localhost:{port} with the model at {self.model_path}...")
      subprocess.check_call(command)
      logger.info("Model server started successfully and running at localhost:{port}")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error running model server: {e}")
      self.clean_up()
      sys.exit(1)

  def _docker_hash(self):
    """Generate a hash of the combined requirements file and Dockefile"""
    with open(self.requirements_file, "r") as f:
      requirements_hash = hashlib.md5(f.read().encode('utf-8')).hexdigest()
    with open(os.path.join(self.model_path, "Dockerfile"), "r") as f:
      dockerfile_hash = hashlib.md5(f.read().encode('utf-8')).hexdigest()

    return hashlib.md5(f"{requirements_hash}{dockerfile_hash}".encode('utf-8')).hexdigest()

  def is_docker_installed(self):
    """Checks if Docker is installed on the system."""
    try:
      logger.info("Checking if Docker is installed...")
      subprocess.run(["docker", "--version"], check=True)
      return True
    except subprocess.CalledProcessError:
      logger.error(
          "Docker is not installed! Please install Docker to run the model in a container.")
      return False

  def build_docker_image(
      self,
      image_name="model_image",
  ):
    """Build the docker image using the Dockerfile in the model directory."""
    try:
      logger.info(f"Building docker image from Dockerfile in {self.model_path}...")

      # since we don't want to copy the model directory into the container, we need to modify the Dockerfile and comment out the COPY instruction
      dockerfile_path = os.path.join(self.model_path, "Dockerfile")
      # Read the Dockerfile
      with open(dockerfile_path, 'r') as file:
        lines = file.readlines()

      # Comment out the COPY instruction that copies the current folder
      modified_lines = []
      for line in lines:
        if 'COPY' in line and '/home/nonroot/main' in line:
          modified_lines.append(f'# {line}')
        elif 'download-checkpoints' in line and '/home/nonroot/main' in line:
          modified_lines.append(f'# {line}')
        else:
          modified_lines.append(line)

      # Create a temporary directory to store the modified Dockerfile
      with tempfile.TemporaryDirectory() as temp_dir:
        temp_dockerfile_path = os.path.join(temp_dir, "Dockerfile.temp")

        # Write the modified Dockerfile to the temporary file
        with open(temp_dockerfile_path, 'w') as file:
          file.writelines(modified_lines)

        # Build the Docker image using the temporary Dockerfile
        subprocess.check_call(
            ['docker', 'build', '-t', image_name, '-f', temp_dockerfile_path, self.model_path])
      logger.info(f"Docker image '{image_name}' built successfully!")
    except subprocess.CalledProcessError as e:
      logger.info(f"Error occurred while building the Docker image: {e}")
      sys.exit(1)

  def docker_image_exists(self, image_name):
    """Check if the Docker image exists."""
    try:
      logger.info(f"Checking if Docker image '{image_name}' exists...")
      subprocess.run(["docker", "inspect", image_name], check=True)
      logger.info(f"Docker image '{image_name}' exists!")
      return True
    except subprocess.CalledProcessError:
      logger.info(f"Docker image '{image_name}' does not exist!")
      return False

  def _gpu_is_available(self):
    """
    Checks if nvidia-smi is available, indicating a GPU is likely accessible.
    """
    return shutil.which("nvidia-smi") is not None

  def run_docker_container(self,
                           image_name,
                           container_name="clarifai-model-container",
                           port=8080,
                           env_vars=None):
    """Runs a Docker container from the specified image."""
    try:
      logger.info(f"Running Docker container '{container_name}' from image '{image_name}'...")
      # Base docker run command
      cmd = ["docker", "run", "--name", container_name, '--rm', "--network", "host"]
      if self._gpu_is_available():
        cmd.extend(["--gpus", "all"])
      # Add volume mappings
      cmd.extend(["-v", f"{self.model_path}:/home/nonroot/main"])
      # Add environment variables
      if env_vars:
        for key, value in env_vars.items():
          cmd.extend(["-e", f"{key}={value}"])
      # Add the image name
      cmd.append(image_name)
      # update the CMD to run the server
      cmd.extend(["--model_path", "/home/nonroot/main", "--grpc", "--port", str(port)])
      # Run the container
      process = subprocess.Popen(cmd,)
      logger.info(
          f"Docker container '{container_name}' is running successfully! access the model at http://localhost:{port}"
      )

      # Function to handle Ctrl+C (SIGINT) gracefully
      def signal_handler(sig, frame):
        logger.info(f"Stopping Docker container '{container_name}'...")
        subprocess.run(["docker", "stop", container_name], check=True)
        process.terminate()
        logger.info(f"Docker container '{container_name}' stopped successfully!")
        time.sleep(1)
        sys.exit(0)

      # Register the signal handler for SIGINT (Ctrl+C)
      signal.signal(signal.SIGINT, signal_handler)
      # Wait for the process to finish (keeps the container running until it's stopped)
      process.wait()
    except subprocess.CalledProcessError as e:
      logger.info(f"Error occurred while running the Docker container: {e}")
      sys.exit(1)
    except Exception as e:
      logger.info(f"Error occurred while running the Docker container: {e}")
      sys.exit(1)

  def test_model_container(self,
                           image_name,
                           container_name="clarifai-model-container",
                           env_vars=None):
    """Test the model inside the Docker container."""
    try:
      logger.info("Testing the model inside the Docker container...")
      # Base docker run command
      cmd = ["docker", "run", "--name", container_name, '--rm', "--network", "host"]
      if self._gpu_is_available():
        cmd.extend(["--gpus", "all"])
      # update the entrypoint for testing the model
      cmd.extend(["--entrypoint", "python"])
      # Add volume mappings
      cmd.extend(["-v", f"{self.model_path}:/home/nonroot/main"])
      # Add environment variables
      if env_vars:
        for key, value in env_vars.items():
          cmd.extend(["-e", f"{key}={value}"])
      # Add the image name
      cmd.append(image_name)
      # update the CMD to test the model inside the container
      cmd.extend([
          "-c",
          "from clarifai.runners.models.model_run_locally import ModelRunLocally; ModelRunLocally('/home/nonroot/main')._run_test()"
      ])
      # Run the container
      subprocess.check_call(cmd)
      logger.info("Model tested successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error testing the model inside the Docker container: {e}")
      sys.exit(1)

  def container_exists(self, container_name="clarifai-model-container"):
    """Check if the Docker container exists."""
    try:
      # Run docker ps -a to list all containers (running and stopped)
      result = subprocess.run(
          ["docker", "ps", "-a", "--filter", f"name={container_name}", "--format", "{{.Names}}"],
          check=True,
          capture_output=True,
          text=True)
      # If the container name is returned, it exists
      if result.stdout.strip() == container_name:
        logger.info(f"Docker container '{container_name}' exists.")
        return True
      else:
        return False
    except subprocess.CalledProcessError as e:
      logger.error(f"Error occurred while checking if container exists: {e}")
      return False

  def stop_docker_container(self, container_name="clarifai-model-container"):
    """Stop the Docker container if it's running."""
    try:
      # Check if the container is running
      result = subprocess.run(
          ["docker", "ps", "--filter", f"name={container_name}", "--format", "{{.Names}}"],
          check=True,
          capture_output=True,
          text=True)
      if result.stdout.strip() == container_name:
        logger.info(f"Docker container '{container_name}' is running. Stopping it...")
        subprocess.run(["docker", "stop", container_name], check=True)
        logger.info(f"Docker container '{container_name}' stopped successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error occurred while stopping the Docker container: {e}")

  def remove_docker_container(self, container_name="clarifai-model-container"):
    """Remove the Docker container."""
    try:
      logger.info(f"Removing Docker container '{container_name}'...")
      subprocess.run(["docker", "rm", container_name], check=True)
      logger.info(f"Docker container '{container_name}' removed successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error occurred while removing the Docker container: {e}")

  def remove_docker_image(self, image_name):
    """Remove the Docker image."""
    try:
      logger.info(f"Removing Docker image '{image_name}'...")
      subprocess.run(["docker", "rmi", image_name], check=True)
      logger.info(f"Docker image '{image_name}' removed successfully!")
    except subprocess.CalledProcessError as e:
      logger.error(f"Error occurred while removing the Docker image: {e}")

  def clean_up(self):
    """Clean up the temporary virtual environment."""
    if os.path.exists(self.temp_dir):
      logger.info("Cleaning up temporary virtual environment...")
      shutil.rmtree(self.temp_dir)


def main(model_path,
         run_model_server=False,
         inside_container=False,
         port=8080,
         keep_env=False,
         keep_image=False,
         skip_dockerfile: bool = False):

  manager = ModelRunLocally(model_path)
  # get whatever stage is in config.yaml to force download now
  # also always write to where upload/build wants to, not the /tmp folder that runtime stage uses
  _, _, _, when, _, _ = manager.builder._validate_config_checkpoints()
  manager.builder.download_checkpoints(
      stage=when, checkpoint_path_override=manager.builder.checkpoint_path)
  if inside_container:
    if not manager.is_docker_installed():
      sys.exit(1)
    if not skip_dockerfile:
      manager.builder.create_dockerfile()
    image_tag = manager._docker_hash()
    model_id = manager.config['model']['id'].lower()
    # must be in lowercase
    image_name = f"{model_id}:{image_tag}"
    container_name = model_id
    if not manager.docker_image_exists(image_name):
      manager.build_docker_image(image_name=image_name)
    try:
      if run_model_server:
        manager.run_docker_container(
            image_name=image_name, container_name=container_name, port=port)
      else:
        manager.test_model_container(image_name=image_name, container_name=container_name)
    finally:
      if manager.container_exists(container_name):
        manager.stop_docker_container(container_name)
        manager.remove_docker_container(container_name=container_name)
      if not keep_image:
        manager.remove_docker_image(image_name)

  else:
    try:
      use_existing_env = manager.create_temp_venv()
      if not use_existing_env:
        manager.install_requirements()
      if run_model_server:
        manager.run_model_server(port)
      else:
        manager.test_model()
    finally:
      if not keep_env:
        manager.clean_up()



from typing import List, Tuple
import numpy as np
from clarifai_grpc.grpc.api import resources_pb2 as respb2


def parse_eval_annotation_classifier(
    eval_metrics: respb2.EvalMetrics) -> Tuple[np.array, np.array, list, List[respb2.Input]]:
  test_set = eval_metrics.test_set
  # get concept ids
  concept_ids = [each.id for each in test_set[0].predicted_concepts]
  concept_ids.sort()
  # get test set
  y_preds = []
  y = []
  inputs = []
  for data in test_set:

    def _to_array(_data):
      cps = [0] * len(concept_ids)
      for each in _data:
        cps[concept_ids.index(each.id)] = each.value
      return np.asarray(cps)

    y_preds.append(_to_array(data.predicted_concepts))
    y.append(_to_array(data.ground_truth_concepts))
    inputs.append(data.input)

  return np.asarray(y), np.asarray(y_preds), concept_ids, inputs


def parse_eval_annotation_detector(eval_metrics: respb2.EvalMetrics,
                                   normalized_box: bool = False,
                                   box_style: str = "xyxy"
                                  ) -> Tuple[np.array, np.array, list, List[respb2.Input]]:
  BOX_STYLES = ["xyxy", "xywh"]
  assert box_style in BOX_STYLES, ValueError(f"Expected box_style in {BOX_STYLES}")

  concept_ids = list(set([each.concept.id for each in eval_metrics.metrics_by_class]))
  concept_ids.sort()

  def _get_box_annot(field, img_height, img_width):
    xyxy_concept_score = []
    for each in field:
      box = each.region_info.bounding_box
      x1 = box.left_col * img_width
      y1 = box.top_row * img_height
      x2 = box.right_col * img_width
      y2 = box.bottom_row * img_height
      score = each.data.concepts[0].value
      concept = each.data.concepts[0].id
      concept_index = concept_ids.index(concept)
      if box_style == "xyxy":
        xyxy_concept_score.append([x1, y1, x2, y2, concept_index, score])
      else:
        w = abs(x1 - x2)
        h = abs(y1 - y2)
        xyxy_concept_score.append([x1, y1, w, h, concept_index, score])

    return np.asarray(xyxy_concept_score)

  inputs = []
  pred_xyxy_concept_score = []
  gt_xyxy_concept_score = []
  for input_data in eval_metrics.test_set:
    _input = input_data.input
    img_height = _input.data.image.image_info.height if not normalized_box else 1.
    img_width = _input.data.image.image_info.height if not normalized_box else 1.
    _pred_xyxy_concept_score = _get_box_annot(
        input_data.predicted_annotation.data.regions, img_height=img_height, img_width=img_width)
    _gt_xyxy_concept_score = _get_box_annot(
        input_data.ground_truth_annotation.data.regions,
        img_height=img_height,
        img_width=img_width)

    pred_xyxy_concept_score.append(_pred_xyxy_concept_score)
    gt_xyxy_concept_score.append(_gt_xyxy_concept_score)
    inputs.append(_input)

  return np.asarray(gt_xyxy_concept_score), np.asarray(
      pred_xyxy_concept_score), concept_ids, inputs


def parse_eval_annotation_detector_coco(eval_metrics: respb2.EvalMetrics,
                                       ) -> Tuple[np.array, np.array, list, List[respb2.Input]]:

  gts, preds, concept_ids, inputs = parse_eval_annotation_detector(
      eval_metrics=eval_metrics, normalized_box=False, box_style="xywh")

  def _make_box_annot(data, input_data, accum_id, is_pred=True):
    img_id = input_data.id
    img_url = input_data.data.image.url
    img_height = input_data.data.image.image_info.height
    img_width = input_data.data.image.image_info.height
    image = {
        "id": img_id,
        "file_name": img_url,
        "width": img_width,
        "height": img_height,
    }
    annotations = []
    for i, d in enumerate(data.tolist()):
      area = d[2] * d[3]
      box = {
          "iscrowd": 0,
          "ignore": 0,
          "image_id": img_id,
          "bbox": d[:4],
          "area": area,
          "segmentation": [],
          "category_id": d[4],
          "id": accum_id + i,
      }
      if is_pred:
        box["score"] = d[5]
      annotations.append(box)

    return image, annotations

  categories = [{
      "supercategory": "none",
      "id": i,
      "name": label
  } for i, label in enumerate(concept_ids)]

  accum_pred_ids = 0
  accum_gt_ids = 0
  pred_images = []
  pred_boxes = []
  gt_images = []
  gt_boxes = []
  for ith, input_proto in enumerate(inputs):
    pred = preds[ith]
    gt = gts[ith]
    pred_img, pred_box = _make_box_annot(
        pred, accum_id=accum_pred_ids, input_data=input_proto, is_pred=True)
    gt_img, gt_box = _make_box_annot(
        gt, accum_id=accum_gt_ids, input_data=input_proto, is_pred=False)

    accum_pred_ids += len(pred)
    pred_images.append(pred_img)
    pred_boxes += pred_box

    accum_gt_ids += len(gt)
    gt_images.append(gt_img)
    gt_boxes += gt_box

  pred_annots = {"images": pred_images, "annotations": pred_boxes, "categories": categories}
  gt_annots = {"images": gt_images, "annotations": gt_boxes, "categories": categories}

  return gt_annots, pred_annots



"""
This is simply the main file for the server that imports ModelRunner implementation
and starts the server.
"""

import argparse
import os
from concurrent import futures

from clarifai_grpc.grpc.api import service_pb2_grpc
from clarifai_protocol.utils.grpc_server import GRPCServer

from clarifai.runners.models.model_builder import ModelBuilder
from clarifai.runners.models.model_runner import ModelRunner
from clarifai.runners.models.model_servicer import ModelServicer
from clarifai.utils.logging import logger


def main():
  parser = argparse.ArgumentParser()
  parser.add_argument(
      '--port',
      type=int,
      default=8000,
      help="The port to host the gRPC server at.",
      choices=range(1024, 65535),
  )
  parser.add_argument(
      '--pool_size',
      type=int,
      default=32,
      help="The number of threads to use for the gRPC server.",
      choices=range(1, 129),
  )  # pylint: disable=range-builtin-not-iterating
  parser.add_argument(
      '--max_queue_size',
      type=int,
      default=10,
      help='Max queue size of requests before we begin to reject requests (default: 10).',
      choices=range(1, 21),
  )  # pylint: disable=range-builtin-not-iterating
  parser.add_argument(
      '--max_msg_length',
      type=int,
      default=1024 * 1024 * 1024,
      help='Max message length of grpc requests (default: 1 GB).',
  )
  parser.add_argument(
      '--enable_tls',
      action='store_true',
      default=False,
      help=
      'Set to true to enable TLS (default: False) since this server is meant for local development only.',
  )
  parser.add_argument(
      '--grpc',
      action='store_true',
      default=False,
      help=
      'Set to true to start the gRPC server (default: False). If set to false, the server will not start and only the runner loop will start to fetch work from the API.',
  )
  parser.add_argument(
      '--model_path',
      type=str,
      required=True,
      help='The path to the model directory that contains implemention of the model.',
  )

  parsed_args = parser.parse_args()

  serve(parsed_args.model_path, parsed_args.port, parsed_args.pool_size,
        parsed_args.max_queue_size, parsed_args.max_msg_length, parsed_args.enable_tls,
        parsed_args.grpc)


def serve(model_path,
          port=8000,
          pool_size=32,
          max_queue_size=10,
          max_msg_length=1024 * 1024 * 1024,
          enable_tls=False,
          grpc=False):

  builder = ModelBuilder(model_path, download_validation_only=True)

  model = builder.create_model_instance()

  # `num_threads` can be set in config.yaml or via the environment variable CLARIFAI_NUM_THREADS="<integer>".
  # Note: The value in config.yaml takes precedence over the environment variable.
  num_threads = builder.config.get("num_threads")

  # Setup the grpc server for local development.
  if grpc:

    # initialize the servicer with the runner so that it gets the predict(), generate(), stream() classes.
    servicer = ModelServicer(model)

    server = GRPCServer(
        futures.ThreadPoolExecutor(
            max_workers=pool_size,
            thread_name_prefix="ServeCalls",
        ),
        max_msg_length,
        max_queue_size,
    )
    server.add_port_to_server('[::]:%s' % port, enable_tls)

    service_pb2_grpc.add_V2Servicer_to_server(servicer, server)
    server.start()
    logger.info("Started server on port %s", port)
    logger.info(f"Access the model at http://localhost:{port}")
    server.wait_for_termination()
  else:  # start the runner with the proper env variables and as a runner protocol.

    # initialize the Runner class. This is what the user implements.
    runner = ModelRunner(
        model=model,
        runner_id=os.environ["CLARIFAI_RUNNER_ID"],
        nodepool_id=os.environ["CLARIFAI_NODEPOOL_ID"],
        compute_cluster_id=os.environ["CLARIFAI_COMPUTE_CLUSTER_ID"],
        base_url=os.environ.get("CLARIFAI_API_BASE", "https://api.clarifai.com"),
        num_parallel_polls=num_threads,
    )
    runner.start()  # start the runner to fetch work from the API.


if __name__ == '__main__':
  main()
